from eflow._hidden.parent_objects import FileOutput
from eflow._hidden.general_objects import DataFrameSnapshot
from eflow.utils.pandas_utils import descr_table,value_counts_table
from eflow._hidden.custom_exceptions import UnsatisfiedRequirments, SnapshotMismatchError
from eflow._hidden.constants import GRAPH_DEFAULTS
from eflow._hidden.parent_objects import DataAnalysis
from eflow.utils.pandas_utils import check_if_feature_exists, generate_meta_data, generate_entropy_table, feature_correlation_table, average_feature_correlation_table
from eflow.utils.sys_utils import dict_to_json_file, pickle_object_to_file, create_dir_structure

import warnings
import random
import numpy as np
from matplotlib import pyplot as plt
import copy
from IPython.display import display
from eflow.utils.pandas_utils import auto_binning
import seaborn as sns
import pandas as pd
from scipy import stats

__author__ = "Eric Cacciavillani"
__copyright__ = "Copyright 2019, eFlow"
__credits__ = ["Eric Cacciavillani"]
__license__ = "MIT"
__maintainer__ = "EricCacciavillani"
__email__ = "eric.cacciavillani@gmail.com"


class FeatureAnalysis(DataAnalysis):

    """
        Analyzes the feature data of a pandas Dataframe object.
        (Ignores null data for displaying data and creates 2d graphics with 2 features.
        In the future I might add 3d graphics with 3 features.)
    """

    def __init__(self,
                 df_features,
                 project_sub_dir="",
                 project_name="Feature Analysis",
                 overwrite_full_path=None,
                 notebook_mode=False):
        """
        Args:

            df_features: DataFrameTypes object from eflow.
                DataFrameTypes object.

            project_sub_dir: string
                Appends to the absolute directory of the output folder

            project_name: string
                Creates a parent or "project" folder in which all sub-directories
                will be inner nested.

            overwrite_full_path: string, None
                Overwrites the path to the parent folder.

            notebook_mode: bool
                If in a python notebook display visualizations in the notebook.
        """

        DataAnalysis.__init__(self,
                              f'{project_sub_dir}/{project_name}',
                              overwrite_full_path)


        self.__df_features = copy.deepcopy(df_features)
        self.__notebook_mode = copy.deepcopy(notebook_mode)

        # Determines if the perform was called to see if we need to re-check
        # the dataframe.
        self.__called_from_perform = False

    def perform_analysis(self,
                         df,
                         dataset_name,
                         target_features=None,
                         display_visuals=True,
                         display_print=True,
                         save_file=True,
                         dataframe_snapshot=True,
                         suppress_runtime_errors=True,
                         figsize=GRAPH_DEFAULTS.FIGSIZE,
                         aggregate_target_feature=True,
                         selected_features=None,
                         extra_tables=True,
                         statistical_analysis_on_aggregates=True):
        """

            Performs all public methods that generate visualizations/insights
            about the data.

        Note:
            Pretty much my personal lazy button for running the entire object
            without specifying any method in particular.

        Args:
            df: pd.Dataframe
                Pandas dataframe object

            dataset_name: string
                The dataset's name; this will create a sub-directory in which your
                generated graph will be inner-nested in.

            target_features: collection of strings or None
                A feature name that both exists in the init df_features
                and the passed dataframe.

                Note
                    If init to 'None' then df_features will try to extract out
                    the target feature.

            display_visuals: bool
                Boolean value to whether or not to display visualizations.

            display_print: bool
                Determines whether or not to print function's embedded print
                statements.

            save_file: bool
                Boolean value to whether or not to save the file.

            dataframe_snapshot: bool
                Boolean value to determine whether or not generate and compare a
                snapshot of the dataframe in the dataset's directory structure.
                Helps ensure that data generated in that directory is correctly
                associated to a dataframe.

            suppress_runtime_errors: bool
                If set to true; when generating any graphs will suppress any runtime
                errors so the program can keep running.

            extra_tables: bool
                When handling two types of features if set to true this will
                    generate any extra tables that might be helpful.
                    Note -
                        These graphics may create duplicates if you already applied
                        an aggregation in 'perform_analysis'

            aggregate_target_feature: bool
                Aggregate the data of the target feature if the data is
                non-continuous data.

                Note
                    In the future I will have this also working with continuous
                    data.

            selected_features: collection object of features
                Will only focus on these selected feature's and will ignore
                the other given features.

            statistical_analysis_on_aggregates: bool
                If set to true then the function 'statistical_analysis_on_aggregates'
                will run; which aggregates the data of the target feature either
                by discrete values or by binning/labeling continuous data.

        Raises:
            If an empty dataframe is passed to this function or if the same
            dataframe is passed to it raise error.
        """
        try:

            self.__called_from_perform = False

            # Raise empty dataframe error
            if df.shape[0] == 0 or np.sum(np.sum(df.isnull()).values) == df.shape[0]:
               raise UnsatisfiedRequirments("Dataframe must contain valid data and "
                                            "not be empty or filled with nulls!")

            # Compare dataframe json file's snapshot to the given dataframe's
            # snapshot
            if dataframe_snapshot:
               df_snapshot = DataFrameSnapshot()
               df_snapshot.check_create_snapshot(df,
                                                 self.__df_features,
                                                 directory_path=self.folder_path,
                                                 sub_dir=f"{dataset_name}/_Extras")

            generate_meta_data(df,
                               self.folder_path,
                               f"{dataset_name}" + "/_Extras")

            generate_entropy_table(df,
                                   self.__df_features,
                                   self.folder_path,
                                   f"{dataset_name}" + "/_Extras/Statistics")

            corr_df = feature_correlation_table(df)

            # Ensures that the correlation dataframe has numerical values
            if len(corr_df):

                self.save_table_as_plot(df=df,
                                        table=corr_df,
                                        show_index=True,
                                        format_float_pos=7,
                                        df_features=self.__df_features,
                                        filename="Correlation Table",
                                        sub_dir=f"{dataset_name}" + "/_Extras/Statistics",
                                        dataframe_snapshot=False,
                                        suppress_runtime_errors=suppress_runtime_errors,
                                        meta_data=False)


                corr_df = average_feature_correlation_table(df)
                self.save_table_as_plot(df=df,
                                        table=corr_df,
                                        show_index=True,
                                        format_float_pos=7,
                                        df_features=self.__df_features,
                                        filename="Average Correlation Table",
                                        sub_dir=f"{dataset_name}" + "/_Extras/Statistics",
                                        dataframe_snapshot=False,
                                        suppress_runtime_errors=suppress_runtime_errors,
                                        meta_data=False)

                # Init color ranking fo plot
                # Ref: http://tinyurl.com/ydgjtmty
                plt.figure(figsize=(13, 10))
                pal = sns.color_palette("GnBu_d", len(corr_df["Average Correlations"]))
                rank = np.array(corr_df["Average Correlations"]).argsort().argsort()
                ax = sns.barplot(y=corr_df.index.tolist(), x=corr_df["Average Correlations"],
                                 palette=np.array(pal[::-1])[rank])
                plt.xticks(rotation=0, fontsize=15)
                plt.yticks(fontsize=15)
                plt.xlabel("Features", fontsize=20, labelpad=20)
                plt.ylabel("Correlation Average", fontsize=20, labelpad=20)
                plt.title("Average Feature Correlation", fontsize=15)
                self.save_plot(df=df,
                               df_features=self.__df_features,
                               filename="Average Correlation Rank Graph",
                               sub_dir=f"{dataset_name}" + "/_Extras/Statistics",
                               dataframe_snapshot=False,
                               suppress_runtime_errors=suppress_runtime_errors,
                               meta_data=False)

                plt.close("all")

                del corr_df

            # Set to true to represent the function call was made with perform
            self.__called_from_perform = True

            if isinstance(target_features,str):
                target_features = {target_features}

            if not target_features:
                target_features = {None}

            if isinstance(target_features,list):
                target_features = set(target_features)

            # Iterate through all target features
            for target_feature in target_features:

                # Iterate through all dataframe features
                for feature_name in df.columns:

                   # Only compare selected features if user specfied features
                   if selected_features and feature_name not in selected_features and feature_name != target_feature:
                       continue

                   # Ignore if the feature is found to be purely null
                   if feature_name in self.__df_features.null_only_features():
                       continue

                   # Ignore datetime features
                   if feature_name in self.__df_features.datetime_features():
                       continue

                   self.analyze_feature(df,
                                        feature_name,
                                        dataset_name,
                                        target_feature=target_feature,
                                        display_visuals=display_visuals,
                                        save_file=save_file,
                                        dataframe_snapshot=dataframe_snapshot,
                                        suppress_runtime_errors=suppress_runtime_errors,
                                        figsize=figsize,
                                        display_print=display_print,
                                        extra_tables=extra_tables)

                   # Aggregate data if target feature exists
                   if target_feature and feature_name == target_feature and aggregate_target_feature:

                       # -----
                       if target_feature in self.__df_features.non_numerical_features() or \
                               target_feature in self.__df_features.bool_features():

                           target_feature_values = df[target_feature].value_counts(sort=False).index.to_list()

                           # Begin aggregation
                           for target_feature_val in target_feature_values:

                               repr_target_feature_val = target_feature_val

                               # Convert to best bool representation of value
                               if target_feature in self.__df_features.bool_features():

                                   try:
                                       repr_target_feature_val = bool(int(repr_target_feature_val))

                                   except ValueError:
                                       continue

                                   except TypeError:
                                       continue

                               # Iterate through all features to generate new graphs for aggregation
                               for f_name in df.columns:

                                   if selected_features and f_name not in selected_features and f_name != target_feature:
                                       continue

                                   if f_name == target_feature:
                                       continue

                                   if display_print:
                                       if repr_target_feature_val:
                                           print(f"Target feature {target_feature} set to {target_feature_val}; also known as {repr_target_feature_val}.")
                                       else:
                                           print(f"Target feature {target_feature} set to {target_feature_val}.")
                                   try:
                                       self.analyze_feature(df[df[target_feature] == target_feature_val],
                                                            f_name,
                                                            dataset_name,
                                                            target_feature=target_feature,
                                                            display_visuals=display_visuals,
                                                            save_file=save_file,
                                                            dataframe_snapshot=dataframe_snapshot,
                                                            suppress_runtime_errors=suppress_runtime_errors,
                                                            figsize=figsize,
                                                            display_print=display_print,
                                                            sub_dir=f"{dataset_name}/{target_feature}/Where {target_feature} = {repr_target_feature_val}/{f_name}",
                                                            extra_tables=False)
                                   except Exception as e:
                                       print(f"Error found on feature {f_name}: {e}")

                # If any missed features are picked up...
                missed_features = set(df.columns) ^ self.__df_features.all_features()
                if len(missed_features) != 0 and display_print:
                    print("Some features were not analyzed by perform analysis!")
                    for feature_name in missed_features:
                        print(f"\t\tFeature:{feature_name}")

            if statistical_analysis_on_aggregates and target_feature:
                self.statistical_analysis_on_aggregates(df,
                                                        target_features,
                                                        dataset_name,
                                                        dataframe_snapshot=False)
        # Ensures that called from perform is turned off
        finally:
            self.__called_from_perform = False


    def analyze_feature(self,
                        df,
                        feature_name,
                        dataset_name,
                        target_feature=None,
                        display_visuals=True,
                        display_print=True,
                        sub_dir=None,
                        save_file=True,
                        dataframe_snapshot=True,
                        suppress_runtime_errors=True,
                        figsize=GRAPH_DEFAULTS.FIGSIZE,
                        extra_tables=True):
        """
            Generate's all graphic's for that given feature and the relationship
            to the target feature.

        Args:
            df: pd.Dataframe
                Pandas DataFrame object

            feature_name: string
                Specified feature column name.

            dataset_name: string
                The dataset's name; this will create a sub-directory in which your
                generated graph will be inner-nested in.

            target_feature: string
                Will create graphics involving this feature with the main
                feature 'feature_name'.

            display_visuals: string
                Boolean value to whether or not to display visualizations.

            display_print: bool
                Determines whether or not to print function's embedded print
                statements.

            sub_dir: string
                Specify the sub directory to append to the pre-defined folder path.

            save_file: bool
                Saves file if set to True; doesn't if set to False.

            dataframe_snapshot: bool
                Boolean value to determine whether or not generate and compare a
                snapshot of the dataframe in the dataset's directory structure.
                Helps ensure that data generated in that directory is correctly
                associated to a dataframe.

            suppress_runtime_errors: bool
                If set to true; when generating any graphs will suppress any runtime
                errors so the program can keep running.

            extra_tables: bool
                When handling two types of features if set to true this will
                generate any extra tables that might be helpful.
                Note -
                    These graphics may create duplicates if you already applied
                    an aggregation in 'perform_analysis'

        Raises:
            Raises error if the json file's snapshot of the given dataframe doesn't
            match the given dataframe.
        """

        # -----
        check_if_feature_exists(df,
                                feature_name)

        colors = self.__get_feature_colors(df,
                                           feature_name)

        # Display colors
        if colors and display_print:
            print(f"Colors:\n{colors}\n")


        # Check if feature exist in df_features and by extension the dataframe
        if target_feature:

            if target_feature not in self.__df_features.all_features():
                raise UnsatisfiedRequirments("Target feature does not exist in pre-defined "
                                             "df_features!")
            if target_feature not in df.columns:
                raise UnsatisfiedRequirments("Target feature does not exist in "
                                             "the dataframe!")

        if feature_name not in self.__df_features.all_features():
            raise UnsatisfiedRequirments(
                "Feature name does not exist in pre-defined "
                "df_features!")
        if feature_name not in df.columns:
            raise UnsatisfiedRequirments("Feature name does not exist in "
                                         "the dataframe!")
        # Generate sub directory structure for plots involving two features
        two_dim_sub_dir = None
        if sub_dir:
            two_dim_sub_dir = sub_dir
        else:
            if target_feature:
                two_dim_sub_dir = f"{dataset_name}/{target_feature}/Two feature analysis/{target_feature} by {feature_name}"

        # -----
        if feature_name in self.__df_features.non_numerical_features() or feature_name in self.__df_features.bool_features():

            # Pie graph's should only have less than or equal to six.
            # (The function can handle ample more than this; just stylistically)
            if len(df[feature_name].value_counts().index) <= 5:
                self.plot_pie_graph(df,
                                    feature_name,
                                    dataset_name=dataset_name,
                                    display_visuals=display_visuals,
                                    sub_dir=sub_dir,
                                    save_file=save_file,
                                    pallete=colors,
                                    dataframe_snapshot=dataframe_snapshot,
                                    suppress_runtime_errors=suppress_runtime_errors,
                                    figsize=figsize,
                                    display_print=display_print)

            # Count plot without colors
            self.plot_count_graph(df,
                                  feature_name,
                                  dataset_name=dataset_name,
                                  display_visuals=display_visuals,
                                  sub_dir=sub_dir,
                                  save_file=save_file,
                                  dataframe_snapshot=dataframe_snapshot,
                                  suppress_runtime_errors=suppress_runtime_errors,
                                  display_print=display_print)

            # Count plot with colors
            if colors:
                self.plot_count_graph(df,
                                      feature_name,
                                      dataset_name=dataset_name,
                                      display_visuals=display_visuals,
                                      sub_dir=sub_dir,
                                      save_file=save_file,
                                      palette=colors,
                                      dataframe_snapshot=dataframe_snapshot,
                                      suppress_runtime_errors=suppress_runtime_errors,
                                      figsize=figsize,
                                      display_print=display_print)

            # Generate value counts table
            self.value_counts_table(df,
                                    feature_name,
                                    dataset_name=dataset_name,
                                    display_visuals=display_visuals,
                                    sub_dir=sub_dir,
                                    save_file=save_file,
                                    dataframe_snapshot=dataframe_snapshot,
                                    suppress_runtime_errors=suppress_runtime_errors,
                                    display_print=display_print)

        # -----
        elif feature_name in self.__df_features.continuous_numerical_features():

            # Plot distance plot graph
            self.plot_distance_graph(df,
                                     feature_name,
                                     dataset_name=dataset_name,
                                     display_visuals=display_visuals,
                                     sub_dir=sub_dir,
                                     save_file=save_file,
                                     dataframe_snapshot=dataframe_snapshot,
                                     suppress_runtime_errors=suppress_runtime_errors,
                                     figsize=figsize,
                                     display_print=display_print)

            # Create description table
            self.descr_table(df,
                             feature_name,
                             dataset_name=dataset_name,
                             display_visuals=display_visuals,
                             sub_dir=sub_dir,
                             save_file=save_file,
                             dataframe_snapshot=dataframe_snapshot,
                             suppress_runtime_errors=suppress_runtime_errors,
                             display_print=display_print)


        if target_feature and feature_name != target_feature:

            # Simplified conditional check for finding type relationship between the two features
            num_features = []
            non_num_features = []

            if target_feature in self.__df_features.continuous_numerical_features():
                num_features.append(target_feature)
            elif target_feature in self.__df_features.datetime_features():
                pass
            elif target_feature not in self.__df_features.continuous_numerical_features():
                non_num_features.append(target_feature)

            if feature_name in self.__df_features.continuous_numerical_features():
                num_features.append(feature_name)
            elif feature_name in self.__df_features.datetime_features():
                pass
            elif feature_name not in self.__df_features.continuous_numerical_features():
                non_num_features.append(feature_name)

            # Two different types of features (numerical and non-numerical)
            if len(num_features) == 1 and len(non_num_features) == 1:

                # Extract out feature name's to better named variables for sanity
                numerical_feature = num_features.pop()
                non_numerical_feature = non_num_features.pop()

                # Generate violin
                self.plot_violin_graph(df,
                                       non_numerical_feature,
                                       dataset_name=dataset_name,
                                       other_feature_name=numerical_feature,
                                       display_visuals=display_visuals,
                                       sub_dir=two_dim_sub_dir,
                                       save_file=save_file,
                                       palette=colors,
                                       dataframe_snapshot=dataframe_snapshot,
                                       suppress_runtime_errors=suppress_runtime_errors,
                                       figsize=figsize,
                                       display_print=display_print)

                # Generate ridge graph
                self.plot_ridge_graph(df,
                                      non_numerical_feature,
                                      dataset_name=dataset_name,
                                      other_feature_name=numerical_feature,
                                      display_visuals=display_visuals,
                                      sub_dir=two_dim_sub_dir,
                                      save_file=save_file,
                                      dataframe_snapshot=dataframe_snapshot,
                                      palette=colors,
                                      suppress_runtime_errors=suppress_runtime_errors,
                                      figsize=figsize,
                                      display_print=display_print)

                # Generate tables based on the aggregation of the non-numerical feature
                if extra_tables:
                    for val in df[non_numerical_feature].unique():

                        if display_print:
                            print(f"Where {non_numerical_feature} = {val}")

                        # Create new sub dir based on the aggregation
                        two_dim_desc_sub_dir = copy.deepcopy(
                            two_dim_sub_dir)
                        if not two_dim_desc_sub_dir:
                            two_dim_desc_sub_dir = ""
                        two_dim_desc_sub_dir += "/" + str(val)

                        # Create new dataframe on aggregated value and check for nans
                        tmp_df = df[df[non_numerical_feature] == val]
                        if np.sum(tmp_df[numerical_feature].isnull()) != \
                                tmp_df.shape[0]:

                            self.descr_table(df=tmp_df,
                                             feature_name=numerical_feature,
                                             dataset_name=dataset_name,
                                             display_visuals=display_visuals,
                                             display_print=display_print,
                                             sub_dir=two_dim_desc_sub_dir,
                                             dataframe_snapshot=False)
                            if display_print:
                                print("\n")

                        del tmp_df

            elif len(non_num_features) == 2:

                # Generate tables based on the aggregation of the non-numerical feature
                if extra_tables:
                    for val in df[feature_name].dropna().unique():

                        if display_print:
                            print(f"Where {feature_name} = {val}")

                        # Create new sub dir based on the aggregation
                        two_dim_desc_sub_dir = copy.deepcopy(
                            two_dim_sub_dir)
                        if not two_dim_desc_sub_dir:
                            two_dim_desc_sub_dir = ""
                        two_dim_desc_sub_dir += "/" + str(val)

                        # Create new dataframe on aggregated value and check for nans
                        tmp_df = df[df[feature_name] == val]
                        if np.sum(tmp_df[target_feature].isnull()) != \
                                tmp_df.shape[0]:
                            self.value_counts_table(df=df[df[feature_name] == val],
                                                    feature_name=target_feature,
                                                    dataset_name=dataset_name,
                                                    display_visuals=display_visuals,
                                                    display_print=display_print,
                                                    sub_dir=two_dim_desc_sub_dir,
                                                    dataframe_snapshot=False)
                            if display_print:
                                print("\n")

                self.group_by_feature_value_count_table(df,
                                                        feature_name,
                                                        dataset_name=dataset_name,
                                                        other_feature_name=target_feature,
                                                        display_visuals=display_visuals,
                                                        sub_dir=two_dim_sub_dir,
                                                        save_file=save_file,
                                                        dataframe_snapshot=dataframe_snapshot,
                                                        suppress_runtime_errors=suppress_runtime_errors,
                                                        display_print=display_print)
                self.plot_multi_bar_graph(df,
                                          feature_name,
                                          dataset_name=dataset_name,
                                          other_feature_name=target_feature,
                                          display_visuals=display_visuals,
                                          sub_dir=two_dim_sub_dir,
                                          save_file=save_file,
                                          dataframe_snapshot=dataframe_snapshot,
                                          suppress_runtime_errors=suppress_runtime_errors,
                                          figsize=figsize,
                                          display_print=display_print,
                                          stacked=False)

                self.plot_multi_bar_graph(df,
                                          feature_name,
                                          dataset_name=dataset_name,
                                          other_feature_name=target_feature,
                                          display_visuals=display_visuals,
                                          sub_dir=two_dim_sub_dir,
                                          save_file=save_file,
                                          dataframe_snapshot=dataframe_snapshot,
                                          suppress_runtime_errors=suppress_runtime_errors,
                                          figsize=figsize,
                                          display_print=display_print,
                                          stacked=True)


            elif len(num_features) == 2:

                # Generate jointplot graph with scatter and kde
                self.plot_jointplot_graph(df,
                                          feature_name,
                                          dataset_name=dataset_name,
                                          other_feature_name=target_feature,
                                          display_visuals=display_visuals,
                                          sub_dir=two_dim_sub_dir,
                                          save_file=save_file,
                                          dataframe_snapshot=dataframe_snapshot,
                                          color=colors,
                                          suppress_runtime_errors=suppress_runtime_errors,
                                          figsize=figsize,
                                          display_print=display_print)

                # Generate jointplot graph with kde
                self.plot_jointplot_graph(df,
                                          feature_name,
                                          dataset_name=dataset_name,
                                          other_feature_name=target_feature,
                                          display_visuals=display_visuals,
                                          sub_dir=two_dim_sub_dir,
                                          save_file=save_file,
                                          dataframe_snapshot=dataframe_snapshot,
                                          color=colors,
                                          suppress_runtime_errors=suppress_runtime_errors,
                                          figsize=figsize,
                                          display_print=display_print,
                                          kind="kde")
        if display_print:
            print("\n\n")


    def plot_distance_graph(self,
                            df,
                            feature_name,
                            dataset_name,
                            display_visuals=True,
                            display_print=True,
                            filename=None,
                            sub_dir=None,
                            save_file=True,
                            dataframe_snapshot=True,
                            suppress_runtime_errors=True,
                            figsize=GRAPH_DEFAULTS.FIGSIZE,
                            bins=None,
                            norm_hist=True,
                            hist=True,
                            kde=True,
                            colors=None,
                            fit=None,
                            fit_kws=None):
        """
            Display a distance plot and save the graph in the correct directory.

        Args:
            df: pd.Dataframe
                Pandas dataframe object

            feature_name: string
                Specified feature column name.

            dataset_name: string
                The dataset's name; this will create a sub-directory in which your
                generated graph will be inner-nested in.

            display_visuals: bool
                Boolean value to whether or not to display visualizations.

            display_print: bool
                Determines whether or not to print function's embedded print
                statements.

            filename: string
                Name to give the file.

            sub_dir: string
                Specify the sub directory to append to the pre-defined folder path.

            save_file: bool
                Boolean value to whether or not to save the file.

            dataframe_snapshot: bool
                Boolean value to determine whether or not generate and compare a
                snapshot of the dataframe in the dataset's directory structure.
                Helps ensure that data generated in that directory is correctly
                associated to a dataframe.

            suppress_runtime_errors: bool
                If set to true; when generating any graphs will suppress any runtime
                errors so the program can keep running.

            figsize: tuple
                The given size of the plot.

            bins: int
                Specification of hist bins, or None to use Freedman-Diaconis rule.

            norm_hist: bool
                If True, the histogram height shows a density rather than a count. This is implied if a KDE or fitted density is plotted.

            hist: bool
                Whether to plot a (normed) histogram.

            kde: bool
                Whether to plot a gaussian kernel density estimate.

            colors : matplotlib color
                Color to plot everything but the fitted curve in.

            fit: functional method
                An object with fit method, returning a tuple that can be passed
                to a pdf method a positional arguments following an grid of
                values to evaluate the pdf on.

            fit_kws : dictionaries, optional
                Keyword arguments for underlying plotting functions.

            Credit to seaborn's author:
            Michael Waskom
            Git username: mwaskom
            Doc Link: http://tinyurl.com/ycco2hok

        Raises:
            Raises error if the feature data is filled with only nulls or if
            the json file's snapshot of the given dataframe doesn't match the
            given dataframe.
        """
        try:

            # -----
            check_if_feature_exists(df,
                                    feature_name)

            # Error check
            if np.sum(df[feature_name].isnull()) == df.shape[0]:
                raise UnsatisfiedRequirments(
                    "Distance plot graph couldn't be generated because " +
                    f"there is only missing data to display in {feature_name}!")

            if display_print:
                print(f"Generating graph for distance plot on {feature_name}")

            feature_values = pd.to_numeric(df[feature_name].dropna(),
                                           errors='coerce').dropna()

            if not len(feature_values):
                raise ValueError(
                    f"The given feature {feature_name} doesn't seem to convert to a numeric vector.")

            # Closes up any past graph info
            plt.close('all')

            # Set foundation graph info
            sns.set(style="whitegrid")
            plt.figure(figsize=figsize)
            plt.title("Distance Plot: " + feature_name)

            # Create seaborn graph
            sns.distplot(feature_values,
                         bins=bins,
                         hist=hist,
                         kde=kde,
                         fit=fit,
                         fit_kws=fit_kws,
                         color=colors,
                         norm_hist=norm_hist)

            # Pass a default name if needed
            if not filename:
                filename = f"Distance plot graph on {feature_name}"

            # Create string sub directory path
            if not sub_dir:
                sub_dir = f"{dataset_name}/{feature_name}"

            # -----
            if save_file:

                if self.__called_from_perform:
                    dataframe_snapshot = False

                self.save_plot(df=df,
                               df_features=self.__df_features,
                               filename=filename,
                               sub_dir=sub_dir,
                               dataframe_snapshot=dataframe_snapshot,
                               suppress_runtime_errors=suppress_runtime_errors,
                               meta_data=not self.__called_from_perform)

            if self.__notebook_mode and display_visuals:
                plt.show()

        except SnapshotMismatchError as e:
            raise e

        except Exception as e:

            if suppress_runtime_errors:
                warnings.warn(
                    f"Distance plot graph throw an error on feature '{feature_name}':\n{str(e)}",
                    RuntimeWarning)
            else:
                raise e

        finally:
            plt.close('all')

    def plot_violin_graph(self,
                          df,
                          feature_name,
                          dataset_name,
                          other_feature_name,
                          display_visuals=True,
                          display_print=True,
                          filename=None,
                          sub_dir=None,
                          save_file=True,
                          dataframe_snapshot=True,
                          suppress_runtime_errors=True,
                          figsize=GRAPH_DEFAULTS.FIGSIZE,
                          order=None,
                          cut=2,
                          scale='area',
                          gridsize=100,
                          width=0.8,
                          palette=None,
                          saturation=0.75):
        """
            Display a violin plot and save the graph in the correct directory.

        Args:
            df: pd.Dataframe
                Pandas dataframe object

            feature_name: string
                Specified feature column name to compare to y.

            dataset_name: string
                The dataset's name; this will create a sub-directory in which your
                generated graph will be inner-nested in.

            other_feature_name: string
                Specified feature column name to compare to x.

            display_visuals: bool
                Boolean value to whether or not to display visualizations.

            filename: string
                Name to give the file.

            sub_dir: string
                Specify the sub directory to append to the pre-defined folder path.

            save_file: bool
                Boolean value to whether or not to save the file.

            dataframe_snapshot: bool
                Boolean value to determine whether or not generate and compare a
                snapshot of the dataframe in the dataset's directory structure.
                Helps ensure that data generated in that directory is correctly
                associated to a dataframe.

            suppress_runtime_errors: bool
                If set to true; when generating any graphs will suppress any runtime
                errors so the program can keep running.

            display_print: bool
                Determines whether or not to print function's embedded print
                statements.

            figsize: tuple
                Size of the given plot.

            order: lists of strings
                Order to plot the categorical levels in, otherwise the levels
                are inferred from the data objects.

            cut: float
                Distance, in units of bandwidth size, to extend the density
                past the extreme datapoints. Set to 0 to limit the violin range
                within the range of the observed data.
                (i.e., to have the same effect as trim=True in ggplot.)

            scale: string
                {area, count, width}
                The method used to scale the width of each violin. If area,
                each violin will have the same area. If count, the width of the
                violins will be scaled by the number of observations in that
                bin. If width, each violin will have the same width.

            gridsize: int
                Number of points in the discrete grid used to compute the kernel density estimate.

            width: float
                Width of a full element when not using hue nesting, or width of
                all the elements for one level of the major grouping variable.

            palette: dict or string
                Colors to use for the different levels of the hue variable.
                Should be something that can be interpreted by color_palette(),
                or a dictionary mapping hue levels to matplotlib colors.

            saturation: float
                Proportion of the original saturation to draw colors at. Large
                patches often look better with slightly desaturated colors, but
                set this to 1 if you want the plot colors to perfectly match
                the input color spec.

            Credit to seaborn's author:
            Michael Waskom
            Git username: mwaskom
            Doc link: http://tinyurl.com/y3hxxzgv

        Raises:
            Raises error if the feature data is filled with only nulls or if
            the json file's snapshot of the given dataframe doesn't match the
            given dataframe.
        """
        try:

            # -----
            check_if_feature_exists(df,
                                    feature_name)

            if other_feature_name:
                check_if_feature_exists(df,
                                        other_feature_name)

            # Error check and create title/part of default file name
            found_features = []
            feature_title = ""
            for feature in (feature_name, other_feature_name):
                if feature:
                    if np.sum(df[feature].isnull()) == df.shape[0]:
                        raise UnsatisfiedRequirments("Count plot graph couldn't be generated because " +
                              f"there is only missing data to display in {feature}!")

                    found_features.append(feature)
                    if len(found_features) == 1:
                        feature_title = f"{feature}"
                    else:
                        feature_title += f" by {feature}"


            if not len(found_features):
                raise UnsatisfiedRequirments("Both x and y feature's are type 'None'. Please pass at least one feature.")

            if np.sum(df[feature_name].isnull()) == df.shape[0]:
                raise UnsatisfiedRequirments(
                    "Violin plot graph couldn't be generated because " +
                    f"there is only missing data to display in {feature_name}!")

            del found_features

            if display_print:
                print("Generating graph violin graph on " + feature_title)

            # Closes up any past graph info
            plt.close('all')

            # Set plot structure
            fig = plt.figure(figsize=figsize)

            plt.title("Violin Plot: " + feature_title)

            feature_values = pd.to_numeric(df[other_feature_name],
                                           errors='coerce').dropna()

            if not len(feature_values):
                raise ValueError("The y feature must contain numerical features.")

            x_values = copy.deepcopy(df[feature_name].dropna())

            # if feature_name in self.__df_features.bool_features():
            #     x_values = pd.to_numeric(x_values,
            #                              errors='ignore')
            #
            #     x_values = ['True' if val == 1 else 'False'
            #                 if val == 0 else val
            #                 for val in x_values]

            # Sort list by x_values
            x_values, feature_values = self.__sort_two_lists(x_values,feature_values)

            warnings.filterwarnings("ignore")

            sns.violinplot(x=x_values,
                           y=feature_values,
                           order=order,
                           cut=cut,
                           scale=scale,
                           gridsize=gridsize,
                           width=width,
                           palette=palette,
                           saturation=saturation)

            warnings.filterwarnings("default")

            # Pass a default name if needed
            if not filename:
                filename = f"Violin plot graph on {feature_title}."

            # Create string sub directory path
            if not sub_dir:
                sub_dir = f"{dataset_name}/{feature_name}"

            # -----
            if save_file:

                if self.__called_from_perform:
                    dataframe_snapshot = False

                self.save_plot(df=df,
                               df_features=self.__df_features,
                               filename=filename,
                               sub_dir=sub_dir,
                               dataframe_snapshot=dataframe_snapshot,
                               suppress_runtime_errors=suppress_runtime_errors,
                               meta_data=not self.__called_from_perform)

            if self.__notebook_mode and display_visuals:
                plt.show()

        except SnapshotMismatchError as e:
            raise e

        except Exception as e:
            warnings.filterwarnings("default")

            if suppress_runtime_errors:
                warnings.warn(
                    f"Plot violin graph an error on feature '{feature_name}':\n{str(e)}",
                    RuntimeWarning)
            else:
                raise e

        finally:
            plt.close('all')
            warnings.filterwarnings("default")

    def plot_count_graph(self,
                         df,
                         feature_name,
                         dataset_name,
                         display_visuals=True,
                         display_print=True,
                         filename=None,
                         sub_dir=None,
                         save_file=True,
                         dataframe_snapshot=True,
                         suppress_runtime_errors=True,
                         figsize=GRAPH_DEFAULTS.FIGSIZE,
                         flip_axis=False,
                         palette="PuBu"):
        """

            Display a barplot with color ranking from a feature's value counts
            from the seaborn libary and save the graph in the correct directory
            structure.

        Args:
            df: pd.Dataframe
                Pandas dataframe object.

            feature_name: string
                Specified feature column name.

            dataset_name: string
                The dataset's name; this will create a sub-directory in which your
                generated graph will be inner-nested in.

            display_visuals: bool
                Boolean value to whether or not to display visualizations.

            display_print: bool
                Determines whether or not to print function's embedded print
                statements.

            filename: string
                Name to give the file.

            sub_dir: string
                Specify the sub directory to append to the pre-defined folder path.

            save_file: bool
                Boolean value to whether or not to save the file.

            dataframe_snapshot: bool
                Boolean value to determine whether or not generate and compare a
                snapshot of the dataframe in the dataset's directory structure.
                Helps ensure that data generated in that directory is correctly
                associated to a dataframe.

            suppress_runtime_errors: bool
                If set to true; when generating any graphs will suppress any runtime
                errors so the program can keep running.

            display_print: bool
                Determines whether or not to print function's embedded print
                statements.

            figsize: tuple
                Size for the given plot.

            flip_axis: bool
                Flip the axis the ploting axis from x to y if set to 'True'.

            palette: dict or string
                String representation of color pallete for ranking from seaborn's pallete.

            Credit to seaborn's author:
            Michael Waskom
            Git username: mwaskom
            Link: http://tinyurl.com/y4pzrgcf

        Raises:
            Raises error if the feature data is filled with only nulls or if
            the json file's snapshot of the given dataframe doesn't match the
            given dataframe.
        """
        try:

            # -----
            check_if_feature_exists(df,
                                    feature_name)

            # Error check
            if np.sum(df[feature_name].isnull()) == df.shape[0]:
                raise UnsatisfiedRequirments(
                    "Count plot graph couldn't be generated because " +
                    f"there is only missing data to display in {feature_name}!")

            if display_print:
                print(f"Count plot graph on {feature_name}")

            # Closes up any past graph info
            plt.close('all')

            # Set graph info
            plt.figure(figsize=figsize)
            sns.set(style="whitegrid")

            value_counts = df[feature_name].dropna().value_counts(sort=True)

            feature_values,counts = value_counts.index, value_counts.values
            del value_counts

            # Find and rank values based on counts for color variation of the graph
            if not palette:
                palette = "PuBu"

            if isinstance(palette,str):
                rank_list = np.argsort(-np.array(counts)).argsort()
                pal = sns.color_palette(palette, len(counts))
                palette = np.array(pal[::-1])[rank_list]

            plt.clf()

            if feature_name in self.__df_features.bool_features():

                i = 0
                for val in feature_values:
                    try:
                        feature_values[i] = float(val)
                    except:
                        pass

                feature_values = [bool(val) if val == 0 or val == 1 else val
                                  for val in feature_values]

            # Flip the graph for visual flare
            if flip_axis:
                ax = sns.barplot(x=counts,
                                 y=feature_values,
                                 palette=palette,
                                 order=feature_values)
            else:
                ax = sns.barplot(x=feature_values,
                                 y=counts,
                                 palette=palette,
                                 order=feature_values)

            # Labels for numerical count of each bar
            for p in ax.patches:
                height = p.get_height()
                ax.text(p.get_x() + p.get_width() / 2.,
                        height + 3,
                        '{:1}'.format(height),
                        ha="center")

            plt.title("Category Count Plot: " + feature_name)

            # Pass a default name if needed
            if not filename:
                filename = f"Count plot graph on {feature_name}"

                if isinstance(palette,np.ndarray):
                    filename += " with count color ranking."

            # Create string sub directory path
            if not sub_dir:
                sub_dir = f"{dataset_name}/{feature_name}"

            # -----
            if save_file:

                if self.__called_from_perform:
                    dataframe_snapshot = False

                self.save_plot(df=df,
                               df_features=self.__df_features,
                               filename=filename,
                               sub_dir=sub_dir,
                               dataframe_snapshot=dataframe_snapshot,
                               suppress_runtime_errors=suppress_runtime_errors,
                               meta_data=not self.__called_from_perform)

            if self.__notebook_mode and display_visuals:
                plt.show()

        except SnapshotMismatchError as e:
            raise e

        except Exception as e:
            if suppress_runtime_errors:
                warnings.warn(
                    f"Plot count graph raised an error on feature '{feature_name}':\n{str(e)}",
                    RuntimeWarning)
            else:
                raise e
        finally:
            plt.close('all')

    def plot_pie_graph(self,
                       df,
                       feature_name,
                       dataset_name,
                       display_visuals=True,
                       display_print=True,
                       filename=None,
                       sub_dir=None,
                       save_file=True,
                       dataframe_snapshot=True,
                       suppress_runtime_errors=True,
                       figsize=GRAPH_DEFAULTS.FIGSIZE,
                       pallete=None):
        """

            Display a pie graph and save the graph in the correct directory.

        Args:
           df:
               Pandas DataFrame object.

           feature_name:
               Specified feature column name.

           dataset_name:
               The dataset's name; this will create a sub-directory in which your
               generated graph will be inner-nested in.

           display_visuals:
               Boolean value to whether or not to display visualizations.

           display_print: bool
                Determines whether or not to print function's embedded print
                statements.

           filename:
               If set to 'None' will default to a pre-defined string;
               unless it is set to an actual filename.

           sub_dir:
               Specify the sub directory to append to the pre-defined folder path.

           save_file:
               Boolean value to whether or not to save the file.

           dataframe_snapshot:
                Boolean value to determine whether or not generate and compare a
                snapshot of the dataframe in the dataset's directory structure.
                Helps ensure that data generated in that directory is correctly
                associated to a dataframe.

           suppress_runtime_errors: bool
                If set to true; when generating any graphs will suppress any runtime
                errors so the program can keep running.

           figsize: tuple
                Size of the plot.

           pallete: dict or string
                Dictionary of all feature values to hex color values.

        Raises:
            Raises error if the feature data is filled with only nulls or if
            the json file's snapshot of the given dataframe doesn't match the
            given dataframe.
        """
        try:

            # -----
            check_if_feature_exists(df,
                                    feature_name)

            if np.sum(df[feature_name].isnull()) == df.shape[0]:
                raise UnsatisfiedRequirments("Pie graph couldn't be generated because " +
                      f"there is only missing data to display in {feature_name}!")

            if display_print:
                print(f"Pie graph on {feature_name}")

            # Closes up any past graph info
            plt.close('all')

            # Find value counts
            value_counts = df[feature_name].dropna().value_counts(sort=False)
            feature_values = value_counts.index.tolist()
            value_count_list = value_counts.values.tolist()

            color_list = None

            plt.figure(figsize=figsize)

            # if feature_name in self.__df_features.bool_features():
            #
            #     i = 0
            #     for val in feature_values:
            #         try:
            #             feature_values[i] = float(val)
            #         except:
            #             pass
            #     feature_values = [bool(val) if val == 0 or val == 1 else val
            #                       for val in feature_values]

            # Sort by feature_values
            feature_values,value_count_list = self.__sort_two_lists(feature_values,
                                                                    value_count_list)

            if isinstance(pallete,dict):
                color_list = []
                for value in tuple(feature_values):
                    try:
                        color_list.append(pallete[value])
                    except KeyError:
                        raise KeyError(f"The given value '{value}' in feature '{feature_name}'"
                                       + " was not found in the passed color dict.")

            # Explode the part of the pie graph that is the maximum of the graph
            explode_array = [0] * len(feature_values)
            explode_array[np.array(value_count_list).argmax()] = .03

            # Plot pie graph
            plt.pie(
                tuple(value_count_list),
                labels=tuple(feature_values),
                shadow=False,
                colors=color_list,
                explode=tuple(explode_array),
                startangle=90,
                autopct='%1.1f%%',
            )

            # Set foundation graph info
            plt.gcf()
            plt.title("Pie Chart: " + feature_name)
            plt.legend(fancybox=True,
                       facecolor='w')

            # Set foundation
            plt.axis('equal')

            # Pass a default name if needed
            if not filename:
                filename = f"Pie graph on {feature_name}"

            # Create string sub directory path
            if not sub_dir:
                sub_dir = f"{dataset_name}/{feature_name}"

            # -----
            if save_file:

                if self.__called_from_perform:
                    dataframe_snapshot = False

                self.save_plot(df=df,
                               df_features=self.__df_features,
                               filename=filename,
                               sub_dir=sub_dir,
                               dataframe_snapshot=dataframe_snapshot,
                               suppress_runtime_errors=suppress_runtime_errors,
                               meta_data=not self.__called_from_perform)


            if self.__notebook_mode and display_visuals:
                plt.show()

        except SnapshotMismatchError as e:
            raise e

        except Exception as e:

            if suppress_runtime_errors:
                warnings.warn(
                    f"Pie graph raised an error on feature '{feature_name}':\n{str(e)}",
                    RuntimeWarning)
            else:
                raise e
        finally:
            plt.close('all')

    def plot_ridge_graph(self,
                         df,
                         feature_name,
                         dataset_name,
                         other_feature_name,
                         display_visuals=True,
                         display_print=True,
                         filename=None,
                         sub_dir=None,
                         save_file=True,
                         dataframe_snapshot=True,
                         suppress_runtime_errors=True,
                         figsize=GRAPH_DEFAULTS.FIGSIZE,
                         palette=None):
        """

            Display a ridge plot and save the graph in the correct directory.

        Args:
            df: pd.Dataframe
                Pandas DataFrame object.

            feature_name: string
                Specified feature column name.

            dataset_name: string
                The dataset's name; this will create a sub-directory in which your
                generated graph will be inner-nested in.

            other_feature_name: string
                Feature to compare to.

            display_visuals: bool
                Boolean value to whether or not to display visualizations.

            display_print: bool
                Determines whether or not to print function's embedded print
                statements.

            filename: string
                If set to 'None' will default to a pre-defined string;
                unless it is set to an actual filename.

            sub_dir: string
                Specify the sub directory to append to the pre-defined folder path.

            save_file: bool
                Boolean value to whether or not to save the file.

            dataframe_snapshot: bool
                Boolean value to determine whether or not generate and compare a
                snapshot of the dataframe in the dataset's directory structure.
                Helps ensure that data generated in that directory is correctly
                associated to a dataframe.

            suppress_runtime_errors: bool
                If set to true; when generating any graphs will suppress any runtime
                errors so the program can keep running.

            display_print: bool
                Determines whether or not to print function's embedded print
                statements.

            figsize: tuple
                Tuple object to represent the plot/image's size.

            palette: dict or string
                Dictionary of all feature values to hex color values.

        Note -
            A large part of this was taken from: http://tinyurl.com/tuou2cn

        Raises:
           Raises error if the feature data is filled with only nulls or if
           the json file's snapshot of the given dataframe doesn't match the
           given dataframe.
        """
        try:

            # -----
            check_if_feature_exists(df,
                                    feature_name)

            # -----
            check_if_feature_exists(df,
                                    other_feature_name)

            # Error check on null data
            if np.sum(df[feature_name].isnull()) == df.shape[0]:
                raise UnsatisfiedRequirments(
                    "Ridge plot graph couldn't be generated because " +
                    f"there is only missing data to display in {feature_name}!")

            if np.sum(df[other_feature_name].isnull()) == df.shape[0]:
                raise UnsatisfiedRequirments(
                    "Ridge plot graph couldn't be generated because " +
                    f"there is only missing data to display in {other_feature_name}!")

            if display_print:
                print(f"Ridge plot graph on {feature_name} by {other_feature_name}.")

            sns.set(style="white",
                    rc={"axes.facecolor": (0, 0, 0, 0)})

            # Temporarily turn off chained assignments
            chained_assignment = pd.options.mode.chained_assignment
            pd.options.mode.chained_assignment = None

            tmp_df = copy.deepcopy(df[[feature_name,other_feature_name]])
            tmp_df[other_feature_name] = pd.to_numeric(tmp_df[other_feature_name],
                                                       errors='coerce')

            # if feature_name in self.__df_features.bool_features():
            #
            #     tmp_df[feature_name] = pd.to_numeric(tmp_df[feature_name],
            #                                          errors='ignore')
            #     tmp_df[feature_name] = ['True' if val == 1 else 'False'
            #                               if val == 0 else val
            #                             for val in tmp_df[feature_name]]

            tmp_df.dropna(inplace=True)

            # Remove any values that only return a single value back
            for val in tmp_df[feature_name].dropna().unique():

                feature_value_counts = tmp_df[other_feature_name][tmp_df[feature_name] == val].dropna().value_counts()

                count_length = len(feature_value_counts.values)
                if len(feature_value_counts.index.to_list()) <= 1 or count_length == 0:
                    tmp_df = tmp_df[tmp_df[feature_name] != val]

                elif count_length == 1 and feature_value_counts.values[0] == 1:
                    tmp_df = tmp_df[tmp_df[feature_name] != val]

            # -----
            # for val in tmp_df[other_feature_name].dropna().unique():
            #
            #     feature_value_counts = tmp_df[feature_name][tmp_df[other_feature_name] == val].dropna().value_counts()
            #
            #     count_length = len(feature_value_counts.values)
            #     if len(feature_value_counts.index.to_list()) <= 1 or count_length == 0:
            #         tmp_df = tmp_df[tmp_df[other_feature_name] != val]
            #     elif count_length == 1 and feature_value_counts.values[0] == 1:
            #         tmp_df = tmp_df[tmp_df[other_feature_name] != val]

            pd.options.mode.chained_assignment = chained_assignment
            del chained_assignment

            # # Sort by dataframe's series of 'feature_name'
            # tmp_df[feature_name], tmp_df[other_feature_name] = self.__sort_two_lists(tmp_df[feature_name],
            #                                                                          tmp_df[other_feature_name])

            # Suppress any warnings that the seaborn's backend raises
            warnings.filterwarnings("ignore")
            sns.set(style="white",
                    rc={"axes.facecolor": (0, 0, 0, 0)})

            if not palette:
                palette = sns.cubehelix_palette(10, rot=-.20, light=.7)

            # Initialize the FacetGrid object
            g = sns.FacetGrid(tmp_df,
                              row=feature_name,
                              hue=feature_name,
                              aspect=15,
                              height=.4,
                              palette=palette)

            # Draw the densities in a few steps
            g.map(sns.kdeplot,
                  other_feature_name,
                  clip_on=False,
                  shade=True,
                  alpha=1,
                  lw=1.5,
                  bw=.2)

            g.map(sns.kdeplot,
                  other_feature_name,
                  clip_on=False,
                  color="w",
                  lw=2,
                  bw=.2)

            g.map(plt.axhline,
                  y=0,
                  lw=2,
                  clip_on=False)

            # Define and use a simple function to label the plot in axes coordinates
            def label(x, color, label):
                ax = plt.gca()
                ax.text(-.1,
                        .2,
                        label,
                        fontweight="bold",
                        color=color,
                        ha="left",
                        va="center",
                        transform=ax.transAxes)

            g.map(label, other_feature_name)

            # Set the subplots to overlap
            g.fig.subplots_adjust(hspace=-.25)

            # Remove axes details that don't play well with overlap
            g.set_titles("")
            g.set(yticks=[])
            g.despine(bottom=True, left=True)

            g.fig.set_size_inches(figsize[0], figsize[1], forward=True)

            g.fig.suptitle(f'{feature_name} by {other_feature_name}')

            warnings.filterwarnings("default")

            # Pass a default name if needed
            if not filename:
                filename = f"Ridge plot graph on {feature_name} by {other_feature_name}"

            # Create string sub directory path
            if not sub_dir:
                sub_dir = f"{dataset_name}/{feature_name}"

            # -----
            if save_file:

                if self.__called_from_perform:
                    dataframe_snapshot = False

                if self.__notebook_mode and display_visuals:
                    plt.show()

                self.save_plot(df=df,
                               df_features=self.__df_features,
                               filename=filename,
                               sub_dir=sub_dir,
                               dataframe_snapshot=dataframe_snapshot,
                               suppress_runtime_errors=suppress_runtime_errors,
                               meta_data=not self.__called_from_perform)

        except SnapshotMismatchError as e:
            raise e

        except Exception as e:
            warnings.filterwarnings("default")

            if suppress_runtime_errors:
                warnings.warn(
                    f"Plot ridge graph raised an error on feature '{feature_name}':\n{str(e)}",
                    RuntimeWarning)
            else:
                raise e
        finally:
            plt.close('all')
            warnings.filterwarnings("default")

    def plot_multi_bar_graph(self,
                             df,
                             feature_name,
                             dataset_name,
                             other_feature_name,
                             display_visuals=True,
                             display_print=True,
                             filename=None,
                             sub_dir=None,
                             save_file=True,
                             dataframe_snapshot=True,
                             suppress_runtime_errors=True,
                             figsize=GRAPH_DEFAULTS.FIGSIZE,
                             colors=None,
                             stacked=False):
        """

            Display a pie graph and save the graph in the correct directory.

        Args:
           df:
               Pandas DataFrame object.

           feature_name:
               Specified feature column name.

           dataset_name:
               The dataset's name; this will create a sub-directory in which your
               generated graph will be inner-nested in.

           display_visuals:
               Boolean value to whether or not to display visualizations.

           display_print: bool
                Determines whether or not to print function's embedded print
                statements.

           filename:
               If set to 'None' will default to a pre-defined string;
               unless it is set to an actual filename.

           sub_dir:
               Specify the sub directory to append to the pre-defined folder path.

           save_file:
               Boolean value to whether or not to save the file.

           dataframe_snapshot:
                Boolean value to determine whether or not generate and compare a
                snapshot of the dataframe in the dataset's directory structure.
                Helps ensure that data generated in that directory is correctly
                associated to a dataframe.

           suppress_runtime_errors: bool
                If set to true; when generating any graphs will suppress any runtime
                errors so the program can keep running.

           figsize: tuple
                Size of the plot.

           colors: dict or string
                Dictionary of all feature values to hex color values.

           stacked: bool
                Determines if the multi bar graph should be stacked or not.

        Raises:
            Raises error if the feature data is filled with only nulls or if
            the json file's snapshot of the given dataframe doesn't match the
            given dataframe.
        """
        try:

            # -----
            check_if_feature_exists(df,
                                    feature_name)

            # -----
            check_if_feature_exists(df,
                                    other_feature_name)


            # Error check on nan data
            if np.sum(df[feature_name].isnull()) == df.shape[0]:
                raise UnsatisfiedRequirments("Multi bar graph on " +
                      f"there is only missing data to display in {feature_name}!")

            if np.sum(df[other_feature_name].isnull()) == df.shape[0]:
                raise UnsatisfiedRequirments("Multi bar graph on " +
                                             f"there is only missing data to display in {other_feature_name}!")

            if display_print:
                print(f"Multi bar graph on {feature_name} by {other_feature_name}")

            # Closes up any past graph info
            plt.close('all')


            if not colors:
                try:
                    colors = [self.__df_features.get_feature_colors(feature_name)[val]
                              for val in list(df.groupby(
                            [other_feature_name, feature_name]).size().unstack().columns)]
                except TypeError:
                    pass
                except KeyError:
                    pass

            g = df.groupby([other_feature_name, feature_name]).size().unstack().plot(
                kind='bar',
                stacked=stacked,
                color=colors,
                figsize=figsize)
            g.legend(loc='upper center',
                     bbox_to_anchor=(1.07, 1),
                     shadow=True,
                     ncol=1)
            sns.set(style="whitegrid")

            plt.title(f"Multi bar graph on {feature_name} by {other_feature_name}")


            # Pass a default name if needed
            if not filename:
                if stacked:
                    filename = f"Multi bar graph on {feature_name} by {other_feature_name}"
                else:
                    filename = f"Multi bar graph stacked on {feature_name} by {other_feature_name}"


            # Create string sub directory path
            if not sub_dir:
                sub_dir = f"{dataset_name}/{feature_name}"

            # -----
            if save_file:

                if self.__called_from_perform:
                    dataframe_snapshot = False

                self.save_plot(df=df,
                               df_features=self.__df_features,
                               filename=filename,
                               sub_dir=sub_dir,
                               dataframe_snapshot=dataframe_snapshot,
                               suppress_runtime_errors=suppress_runtime_errors,
                               meta_data=not self.__called_from_perform)

            if self.__notebook_mode and display_visuals:
                plt.show()

        except SnapshotMismatchError as e:
            raise e

        except Exception as e:

            if suppress_runtime_errors:
                warnings.warn(
                    f"Multi bar raised an error on feature '{feature_name}':\n{str(e)}",
                    RuntimeWarning)
            else:
                raise e
        finally:
            plt.close('all')

    def statistical_analysis_on_aggregates(self,
                                           df,
                                           target_features,
                                           dataset_name,
                                           dataframe_snapshot=True):
        """

            Aggregates the data of the target feature either by discrete values
            or by binning/labeling continuous data.

        Args:
           df: pd.Dataframe
               Pandas DataFrame object.

           target_features: list of string
               Specified target features.

           dataset_name: string
               The dataset's name; this will create a sub-directory in which your
               generated graph will be inner-nested in.

           dataframe_snapshot: bool
                Boolean value to determine whether or not generate and compare a
                snapshot of the dataframe in the dataset's directory structure.
                Helps ensure that data generated in that directory is correctly
                associated to a dataframe.

        Note:
            This function has a lot going on and it's infancy so I am going to
            purposely not give it suppress_runtime_errors so people will find
            problems with it and report it to me.

        Raises:
            Raises error if the feature data is filled with only nulls or if
            the json file's snapshot of the given dataframe doesn't match the
            given dataframe.
        """

        if not self.__called_from_perform and dataframe_snapshot:
            df_snapshot = DataFrameSnapshot()
            df_snapshot.check_create_snapshot(df,
                                              self.__df_features,
                                              directory_path=self.folder_path + f"/{dataset_name}",
                                              sub_dir=f"{dataset_name}/_Extras")

        feature_stats_dict = dict()


        # Convert to list of the given string
        if isinstance(target_features, str):
            target_features = [target_features]

        # Iterate through all target features features
        for feature_name in target_features:

            if feature_name:
                check_if_feature_exists(df,
                                        feature_name)
            else:
                continue

            # Generate bins and labels for continuous numerical data
            bins = None
            labels = None
            if feature_name in self.__df_features.continuous_numerical_features():
                bins, labels = auto_binning(df,
                                            self.__df_features,
                                            feature_name)

            # Labels and bins will act as feature values for aggregation
            if labels:
                feature_values = copy.deepcopy(labels)
            else:
                feature_values = list(df[feature_name].sort_values(
                    ascending=True).dropna().unique())

            # Label to compare to all the data without any aggregations
            feature_values.append("All")

            # Copy feature values to remove one given value from the feature list at a time
            other_feature_values = copy.deepcopy(feature_values)

            # Create target feature dict
            feature_stats_dict[feature_name] = dict()
            feature_val_count = -1
            feature_pvalues = dict()

            # Store all pvalues found for every feature
            for column in df.columns:

                if feature_name == column:
                    continue

                feature_pvalues[column] = dict()

            # Iterate through remaining features
            for main_feature_val in feature_values:

                # Ignore "All" for the main feature val since it doesn't actually exist
                if main_feature_val == "All":
                    continue

                # Don't want repeats or compare same subsets
                other_feature_values.remove(main_feature_val)
                feature_val_count += 1

                # Create series objects based on the main feature val compared to the other_feature_val
                other_feature_val_count = 0
                for other_feature_val in other_feature_values:
                    feature_stats_dict[feature_name][
                        f"{main_feature_val} -> {other_feature_val}"] = dict()

                    # Generate series object based on bins/discrete values
                    other_feature_val_count += 1
                    for iterate_feature_name in df.columns:

                        # -----
                        if iterate_feature_name == feature_name:
                            continue

                        # Create bool array for series object(left)
                        if labels:
                            bool_array = (df[feature_name] <= bins[
                                feature_val_count + 1]) & (
                                                     df[feature_name] > bins[
                                                 feature_val_count])
                        else:
                            bool_array = df[feature_name] == main_feature_val

                        tmp_series_a = df[bool_array][
                            iterate_feature_name].dropna()
                        del bool_array

                        # Create bool array for series object(right)
                        if other_feature_val == "All":
                            bool_array = [True for _ in range(0, df.shape[0])]
                        else:
                            if labels:
                                bool_array = (df[feature_name] <= bins[
                                    feature_val_count + other_feature_val_count + 1]) & (
                                                     df[feature_name] > bins[
                                                 feature_val_count + other_feature_val_count])

                            else:
                                bool_array = df[
                                                 feature_name] == other_feature_val

                        tmp_series_b = df[bool_array][
                            iterate_feature_name].dropna()
                        del bool_array

                        # Extract out pvalue/statistic based on series data
                        if len(tmp_series_a) == 0 or len(tmp_series_b) == 0:
                            pvalue = "NaN"
                            statistic = "NaN"
                        else:
                            ks_2samp = stats.ks_2samp(tmp_series_a,
                                                      tmp_series_b)
                            pvalue = float(ks_2samp.pvalue)
                            statistic = float(ks_2samp.statistic)

                        # Init pvalue/statistic to proper values
                        feature_stats_dict[feature_name][
                            f"{main_feature_val} -> {other_feature_val}"][
                            iterate_feature_name] = dict()
                        feature_stats_dict[feature_name][
                            f"{main_feature_val} -> {other_feature_val}"][
                            iterate_feature_name][
                            "Kolmogorov-Smirnov statistic"] = dict()
                        feature_stats_dict[feature_name][
                            f"{main_feature_val} -> {other_feature_val}"][
                            iterate_feature_name][
                            "Kolmogorov-Smirnov statistic"]["P-Value"] = pvalue

                        feature_stats_dict[feature_name][
                            f"{main_feature_val} -> {other_feature_val}"][
                            iterate_feature_name][
                            "Kolmogorov-Smirnov statistic"][
                            "Statistic"] = statistic

                        # Don't add to list
                        if pvalue == "NaN":
                            continue

                        # Init dict/list if it doesn't exist
                        if "Kolmogorov-Smirnov statistic" not in \
                                feature_pvalues[iterate_feature_name]:
                            feature_pvalues[iterate_feature_name][
                                "Kolmogorov-Smirnov statistic"] = dict()
                            feature_pvalues[iterate_feature_name][
                                "Kolmogorov-Smirnov statistic"][
                                "All pvalues"] = []

                        # Append new pvalue
                        feature_pvalues[iterate_feature_name][
                            "Kolmogorov-Smirnov statistic"][
                            "All pvalues"].append(pvalue)

            # Generate summary data of pvalues
            for column in df.columns:
                if column == feature_name or len(feature_pvalues[column].keys()) == 0:
                    continue

                else:
                    if column in feature_pvalues:
                        feature_pvalues[column][
                            "Kolmogorov-Smirnov statistic"][
                            "All pvalues"].sort()

                        # Only create summary if the series is at least the of 2
                        if len(feature_pvalues[column][
                                   "Kolmogorov-Smirnov statistic"][
                                   "All pvalues"]) >= 2:

                            feature_pvalues[column][
                                "Kolmogorov-Smirnov statistic"
                            ]["Pvalues Summary"] = descr_table(
                                pd.DataFrame({column: feature_pvalues[column][
                                    "Kolmogorov-Smirnov statistic"][
                                    "All pvalues"]}),
                                column).to_dict()[column]
                        # Init to an empty dict
                        else:
                            feature_pvalues[column][
                                "Kolmogorov-Smirnov statistic"][
                                "Pvalues Summary"] = {}

            feature_stats_dict[feature_name]["P-Values"] = feature_pvalues
        # End target feature loop

        # Generate directories
        create_dir_structure(self.folder_path + dataset_name,
                             "_Extras/Statistics/Accept Null Hypothesis")
        create_dir_structure(self.folder_path + dataset_name,
                             "_Extras/Statistics/Reject Null Hypothesis")

        # Create json file
        dict_to_json_file(feature_stats_dict,
                          self.folder_path + dataset_name + "/_Extras/Statistics",
                          "Statistics on target features")

        stat_methods_dict = dict()

        for main_feature, relationship_dict in feature_stats_dict.items():
            for _, stats_on_features in relationship_dict.items():
                for iterate_feature_name, stats_method_dict in stats_on_features.items():
                    for stats_method, stats_dict in stats_method_dict.items():
                        if "All pvalues" in stats_dict:

                            if stats_method not in stat_methods_dict:
                                stat_methods_dict[stats_method] = pd.DataFrame()

                            stats_dict = copy.deepcopy(stats_dict)

                            for k,v in stats_dict.items():
                                if v == "NaN":
                                    stats_dict[k] = [np.nan]
                                else:
                                    stats_dict[k] = [v]

                            if len(stats_dict['Pvalues Summary'][0]) > 0:
                                tmp_stats_df = pd.DataFrame.from_dict(stats_dict["Pvalues Summary"])[["mean","std","var"]]
                                tmp_stats_df.index = [f"{main_feature} compared to {iterate_feature_name}"]
                                stat_methods_dict[stats_method] = stat_methods_dict[stats_method].append(tmp_stats_df,
                                                                                           ignore_index=False)
        for stats_method in stat_methods_dict:
            if stat_methods_dict[stats_method].shape[0]:
                stat_methods_dict[stats_method].sort_values(
                    by=["mean", "std", "var"],
                    ascending=True,
                    inplace=True)

        pickle_object_to_file(stat_methods_dict,
                              self.folder_path + dataset_name + "/_Extras/Statistics",
                              "Stat methods of features dataframes",
                              remove_file_extension=False)

        # Generate multiple json files based on the following pvalues
        for accept_null_plvalue in [.01, .05, .1, .101, .2, .3, .4, .5, .6, .7,
                                    .8, .9, 1]:
            json_dict = copy.deepcopy(feature_stats_dict)
            tmp_dict = copy.deepcopy(feature_stats_dict)
            for main_feature, relationship_dict in tmp_dict.items():
                for relationship_string, stats_on_features in relationship_dict.items():
                    for iterate_feature_name, stats_method_dict in stats_on_features.items():
                        for stats_method, stats_dict in stats_method_dict.items():

                            # Not a relationship string; Re-access pvalue list and summary
                            if relationship_string == "P-Values":

                                if "P-Values" not in json_dict[
                                    main_feature] or iterate_feature_name not in \
                                        json_dict[main_feature]["P-Values"]:
                                    break

                                filter_pvalues = np.asarray(
                                    json_dict[main_feature]["P-Values"][
                                        iterate_feature_name][stats_method][
                                        "All pvalues"])

                                if accept_null_plvalue <= .1:
                                    filter_pvalues = filter_pvalues[
                                        filter_pvalues <= accept_null_plvalue]
                                else:
                                    filter_pvalues = filter_pvalues[
                                        filter_pvalues >= accept_null_plvalue]

                                json_dict[main_feature]["P-Values"][
                                    iterate_feature_name][stats_method][
                                    "All pvalues"] = list(filter_pvalues)

                                if len(filter_pvalues) >= 2:
                                    json_dict[main_feature][
                                        "P-Values"][iterate_feature_name][
                                        stats_method]["Pvalues Summary"] = \
                                    descr_table(
                                        pd.DataFrame(
                                            {iterate_feature_name: list(
                                                filter_pvalues)}),
                                        iterate_feature_name).to_dict()[
                                        iterate_feature_name]
                                else:
                                    json_dict[main_feature][
                                        "P-Values"][iterate_feature_name][
                                        stats_method]["Pvalues Summary"] = {}

                                break

                            pvalue = stats_dict["P-Value"]
                            if accept_null_plvalue <= .1:
                                if pvalue == "NaN" or pvalue > accept_null_plvalue:
                                    del json_dict[main_feature][
                                        relationship_string][
                                        iterate_feature_name]
                            else:
                                if pvalue == "NaN" or pvalue < accept_null_plvalue:
                                    del json_dict[main_feature][
                                        relationship_string][
                                        iterate_feature_name]

            # Push to accept or reject null hypothesis folder
            if accept_null_plvalue <= .1:
                dict_to_json_file(json_dict,
                                  self.folder_path + dataset_name + "/_Extras/Statistics/Accept Null Hypothesis",
                                  f"Accept Null Hypothesis on target features where pvalue <= {accept_null_plvalue}",
                                  remove_file_extension=False)
            else:
                dict_to_json_file(json_dict,
                                  self.folder_path + dataset_name + "/_Extras/Statistics/Reject Null Hypothesis",
                                  f"Reject Null Hypothesis on target features where pvalue >= {accept_null_plvalue}",
                                  remove_file_extension=False)

    def plot_jointplot_graph(self,
                             df,
                             feature_name,
                             dataset_name,
                             other_feature_name,
                             display_visuals=True,
                             display_print=True,
                             filename=None,
                             sub_dir=None,
                             save_file=True,
                             dataframe_snapshot=True,
                             suppress_runtime_errors=True,
                             figsize=GRAPH_DEFAULTS.FIGSIZE,
                             color=None,
                             kind="scatter and kde",
                             ratio=5):

        """

            Display a ridge plot and save the graph in the correct directory.

        Args:
            df: pd.Dataframe
                Pandas DataFrame object.

            feature_name: string
                Specified feature column name.

            dataset_name: string
                The dataset's name; this will create a sub-directory in which your
                generated graph will be inner-nested in.

            other_feature_name: string
                Feature to compare to.

            display_visuals: bool
                Boolean value to whether or not to display visualizations.

            display_print: bool
                Determines whether or not to print function's embedded print
                statements.

            filename: string
                If set to 'None' will default to a pre-defined string;
                unless it is set to an actual filename.

            sub_dir: string
                Specify the sub directory to append to the pre-defined folder path.

            save_file: bool
                Boolean value to whether or not to save the file.

            dataframe_snapshot: bool
                Boolean value to determine whether or not generate and compare a
                snapshot of the dataframe in the dataset's directory structure.
                Helps ensure that data generated in that directory is correctly
                associated to a dataframe.

            suppress_runtime_errors: bool
                If set to true; when generating any graphs will suppress any runtime
                errors so the program can keep running.

            display_print: bool
                Determines whether or not to print function's embedded print
                statements.

            figsize: tuple
                Tuple object to represent the plot/image's size. Because joinplot
                only accepts a single value for the figure; we just pull the
                greatest of the two values.

            color: string
                Seaborn/maplotlib color/hex color for representing the graph

            kind: string (scatter,reg,resid,kde,hex,scatter and kde)
                Kind of plot to draw.

            ratio:
                Ratio of joint axes height to marginal axes height.
                (Determines distplot like plots dimensions.)

            Credit to seaborn's author:
            Michael Waskom
            Git username: mwaskom
            Link: http://tinyurl.com/v9pxsoy

        Raises:
           Raises error if the feature data is filled with only nulls or if
           the json file's snapshot of the given dataframe doesn't match the
           given dataframe.
        """


        try:

            # -----
            check_if_feature_exists(df,
                                    feature_name)

            check_if_feature_exists(df,
                                    other_feature_name)

            # Error check on null data
            if np.sum(df[feature_name].isnull()) == df.shape[0]:
                raise UnsatisfiedRequirments(
                    "Jointplot plot graph couldn't be generated because " +
                    f"there is only missing data to display in {feature_name}!")

            if np.sum(df[other_feature_name].isnull()) == df.shape[0]:
                raise UnsatisfiedRequirments(
                    "Jointplot plot graph couldn't be generated because " +
                    f"there is only missing data to display in {other_feature_name}!")

            if display_print:
                print(f"Generating jointplot graph on {feature_name} by {other_feature_name}")

            # Closes up any past graph info
            plt.close('all')

            if figsize[0] < figsize[1]:
                height = figsize[0]
            else:
                height = figsize[1]

            tmp_df = copy.deepcopy(df[[feature_name,other_feature_name]])
            tmp_df.dropna()

            if not kind:
                kind = "scatter"

            warnings.filterwarnings("ignore")
            if kind == "scatter and kde":
                g = sns.jointplot(feature_name,
                                  other_feature_name,
                                  data=tmp_df,
                                  kind="scatter",
                                  color=color,
                                  ratio=ratio,
                                  height=height).plot_joint(sns.kdeplot, zorder=0,
                                                            n_levels=6)
            else:
                g = sns.jointplot(feature_name,
                                  other_feature_name,
                                  data=tmp_df,
                                  kind=kind,
                                  color=color,
                                  ratio=ratio,
                                  height=height)

            warnings.filterwarnings("default")

            plt.subplots_adjust(top=0.93)
            g.fig.suptitle("Jointplot: " + f"{feature_name} by {other_feature_name}")

            # Pass a default name if needed
            if not filename:
                filename = f"Jointplot plot graph for {feature_name} by {other_feature_name} using {kind}"

            # Create string sub directory path
            if not sub_dir:
                sub_dir = f"{dataset_name}/{feature_name}"

            # -----
            if save_file:

                if self.__called_from_perform:
                    dataframe_snapshot = False

                self.save_plot(df=df,
                               df_features=self.__df_features,
                               filename=filename,
                               sub_dir=sub_dir,
                               dataframe_snapshot=dataframe_snapshot,
                               suppress_runtime_errors=suppress_runtime_errors,
                               meta_data=not self.__called_from_perform)

            if self.__notebook_mode and display_visuals:
                plt.show()

        except SnapshotMismatchError as e:
            raise e

        except Exception as e:
            warnings.filterwarnings("default")

            if suppress_runtime_errors:
                warnings.warn(
                    f"Joinplot plot graph an error on feature '{feature_name}':\n{str(e)}",
                    RuntimeWarning)
            else:
                raise e

        finally:
            plt.close('all')
            warnings.filterwarnings("default")


    def value_counts_table(self,
                           df,
                           feature_name,
                           dataset_name,
                           display_visuals=True,
                           display_print=True,
                           filename=None,
                           sub_dir=None,
                           save_file=True,
                           dataframe_snapshot=True,
                           suppress_runtime_errors=True):
        """

            Creates a value counts table of the features given data.

            Note
                Creates a png of the table.

        Args:
            df: pd.Dataframe
                Pandas DataFrame object

            feature_name: string
                Specified feature column name.

            dataset_name: string
                The dataset's name; this will create a sub-directory in which your
                generated graph will be inner-nested in.

            display_visuals: bool
                Boolean value to whether or not to display visualizations.

            display_print: bool
                Determines whether or not to print function's embedded print
                statements.

            filename: string
                If set to 'None' will default to a pre-defined string;
                unless it is set to an actual filename.

            sub_dir: string
                Specify the sub directory to append to the pre-defined folder path.

            save_file: bool
                Saves file if set to True; doesn't if set to False.

            dataframe_snapshot: bool
                Boolean value to determine whether or not generate and compare a
                snapshot of the dataframe in the dataset's directory structure.
                Helps ensure that data generated in that directory is correctly
                associated to a dataframe.

            suppress_runtime_errors: bool
                If set to true; when generating any graphs will suppress any runtime
                errors so the program can keep running.


            Creates/Saves a pandas dataframe of value counts of a dataframe.

            Note -
                Creates a png of the table.

        Raises:
            Raises error if the feature data is filled with only nulls or if
            the json file's snapshot of the given dataframe doesn't match the
            given dataframe.
        """
        try:

            # -----
            check_if_feature_exists(df,
                                    feature_name)

            # Check if feature has only null data
            if np.sum(df[feature_name].isnull()) == df.shape[0]:
                raise UnsatisfiedRequirments("Values count table couldn't be generated because " +
                                             f"there is only missing data to display in {feature_name}!")


            if display_print:
                print(f"Creating value counts table for feature {feature_name}.")

            # -----
            val_counts_df = value_counts_table(df,
                                               feature_name)

            if self.__notebook_mode:
                if display_visuals:
                    display(val_counts_df)
            else:
                if display_visuals:
                    print(val_counts_df)

            # Pass a default name if needed
            if not filename:
                filename = f"{feature_name} Value Counts Table"

            # Create string sub directory path
            if not sub_dir:
                sub_dir = f"{dataset_name}/{feature_name}"

            if save_file:

                if self.__called_from_perform:
                    dataframe_snapshot = False

                self.save_table_as_plot(df=df,
                                        df_features=self.__df_features,
                                        filename=filename,
                                        sub_dir=sub_dir,
                                        dataframe_snapshot=dataframe_snapshot,
                                        suppress_runtime_errors=suppress_runtime_errors,
                                        table=val_counts_df,
                                        show_index=True,
                                        meta_data=not self.__called_from_perform)

        except SnapshotMismatchError as e:
            raise e

        except Exception as e:

            if suppress_runtime_errors:
                warnings.warn(
                    f"Value count table raised an error on feature '{feature_name}':\n{str(e)}",
                    RuntimeWarning)
            else:
                raise e
        finally:
            plt.close('all')


    def descr_table(self,
                    df,
                    feature_name,
                    dataset_name,
                    display_visuals=True,
                    display_print=True,
                    filename=None,
                    sub_dir=None,
                    save_file=True,
                    dataframe_snapshot=True,
                    suppress_runtime_errors=True):
        """

            Creates/Saves a pandas dataframe of a feature's numerical data.
            Standard deviation, mean, Q1-Q5, median, variance, etc.

            Note
                Creates a png of the table.

        Args:
            df: pd.Dataframe
                Pandas DataFrame object

            feature_name: string
                Specified feature column name.

            dataset_name: string
                The dataset's name; this will create a sub-directory in which your
                generated graph will be inner-nested in.

            display_visuals: bool
                Boolean value to whether or not to display visualizations.

            display_print: bool
                Determines whether or not to print function's embedded print
                statements.

            filename: string
                If set to 'None' will default to a pre-defined string;
                unless it is set to an actual filename.

            sub_dir: string
                Specify the sub directory to append to the pre-defined folder path.

            save_file: bool
                Saves file if set to True; doesn't if set to False.

            dataframe_snapshot: bool
                Boolean value to determine whether or not generate and compare a
                snapshot of the dataframe in the dataset's directory structure.
                Helps ensure that data generated in that directory is correctly
                associated to a dataframe.

            suppress_runtime_errors: bool
                If set to true; when generating any graphs will suppress any runtime
                errors so the program can keep running.

        Raises:
            Raises error if the feature data is filled with only nulls or if
            the json file's snapshot of the given dataframe doesn't match the
            given dataframe.
        """
        try:

            # -----
            check_if_feature_exists(df,
                                    feature_name)

            # Check if dataframe has only null data
            if np.sum(df[feature_name].isnull()) == df.shape[0]:
                raise UnsatisfiedRequirments("Descr table couldn't be generated because " +
                      f"there is only missing data to display in {feature_name}!")

            if display_print:
                print(f"Creating data description table for {feature_name}")

            desc_df = descr_table(df,
                                  feature_name,
                                  to_numeric=True)

            if self.__notebook_mode:
                if display_visuals:
                    display(desc_df)
            else:
                if display_visuals:
                    print(desc_df)

            # Pass a default name if needed
            if not filename:
                filename = f"{feature_name} Description Table"

            # Create string sub directory path
            if not sub_dir:
                sub_dir = f"{dataset_name}/{feature_name}"

            if save_file:

                if self.__called_from_perform:
                    dataframe_snapshot = False

                self.save_table_as_plot(df=df,
                                        df_features=self.__df_features,
                                        filename=filename,
                                        sub_dir=sub_dir,
                                        dataframe_snapshot=dataframe_snapshot,
                                        suppress_runtime_errors=suppress_runtime_errors,
                                        table=desc_df,
                                        meta_data=not self.__called_from_perform,
                                        show_index=True)

        except SnapshotMismatchError as e:
            raise e

        except Exception as e:

            if suppress_runtime_errors:
                warnings.warn(
                    f"Descr table raised an error on feature '{feature_name}':\n{str(e)}",
                    RuntimeWarning)
            else:
                raise e

        finally:
            plt.close('all')

    def group_by_feature_value_count_table(self,
                                           df,
                                           feature_name,
                                           dataset_name,
                                           other_feature_name,
                                           display_visuals=True,
                                           display_print=True,
                                           filename=None,
                                           sub_dir=None,
                                           save_file=True,
                                           dataframe_snapshot=True,
                                           suppress_runtime_errors=True):
        """

            Creates/Saves a pandas dataframe of features and their found types
            in the dataframe.

            Note
                Creates a png of the table.

        Args:
            df: pd.Dataframe
                Pandas DataFrame object

            feature_name: string
                Specified feature column name.

            dataset_name: string
                The dataset's name; this will create a sub-directory in which your
                generated graph will be inner-nested in.

            other_feature_name: string
                Feature to compare to.

            display_visuals: bool
                Boolean value to whether or not to display visualizations.

            display_print: bool
                Determines whether or not to print function's embedded print
                statements.

            filename: string
                If set to 'None' will default to a pre-defined string;
                unless it is set to an actual filename.

            sub_dir: string
                Specify the sub directory to append to the pre-defined folder path.

            save_file: bool
                Saves file if set to True; doesn't if set to False.

            dataframe_snapshot: bool
                Boolean value to determine whether or not generate and compare a
                snapshot of the dataframe in the dataset's directory structure.
                Helps ensure that data generated in that directory is correctly
                associated to a dataframe.

            suppress_runtime_errors: bool
                If set to true; when generating any graphs will suppress any runtime
                errors so the program can keep running.

        Raises:
            Raises error if the feature data is filled with only nulls or if
            the json file's snapshot of the given dataframe doesn't match the
            given dataframe.
        """
        try:

            # -----
            check_if_feature_exists(df,
                                    feature_name)

            # Check if dataframe has only null data
            if np.sum(df[feature_name].isnull()) == df.shape[0]:
                raise UnsatisfiedRequirments("Count plot graph couldn't be generated because " +
                      f"there is only missing data to display in {feature_name}!")

            if np.sum(df[other_feature_name].isnull()) == df.shape[0]:
                raise UnsatisfiedRequirments("Count plot graph couldn't be generated because " +
                      f"there is only missing data to display in {other_feature_name}!")

            if display_print:
                print(f"Creating group by {feature_name} and {other_feature_name} Table")

            tmp_df = copy.deepcopy(df[[feature_name, other_feature_name]])
            tmp_df = tmp_df.groupby([feature_name, other_feature_name]).size().to_frame()
            tmp_df.columns = ["Counts"]

            if self.__notebook_mode:
                if display_visuals:
                    display(tmp_df)
            else:
                if display_visuals:
                    print(tmp_df)

            # Pass a default name if needed
            if not filename:
                filename = f"Group by {feature_name} and {other_feature_name} Table"

            # Create string sub directory path
            if not sub_dir:
                sub_dir = f"{dataset_name}/{feature_name}"

            tmp_df.sort_values(by=["Counts"],
                               ascending=False,
                               inplace=True)

            if save_file:

                if self.__called_from_perform:
                    dataframe_snapshot = False

                self.save_table_as_plot(df=tmp_df,
                                        df_features=self.__df_features,
                                        filename=filename,
                                        sub_dir=sub_dir,
                                        dataframe_snapshot=dataframe_snapshot,
                                        suppress_runtime_errors=suppress_runtime_errors,
                                        table=tmp_df,
                                        show_index=True,
                                        meta_data=not self.__called_from_perform)

        except SnapshotMismatchError as e:
            raise e

        except Exception as e:

            if suppress_runtime_errors:
                warnings.warn(
                    f"Group by table raised an error on feature '{feature_name}' by '{other_feature_name}':\n{str(e)}",
                    RuntimeWarning)
            else:
                raise e

        finally:
            plt.close('all')


    def __get_feature_colors(self,
                             df,
                             feature_name):
        """

            Creates a dict object of all possible feature values with their
            associated colors.

            Note
                Any unknown feature values that aren't declared
                by df_features are given a default color from the constants
                section of the project. Goes up to 20 different colors unitl
                colors is init to None.

        Args:
            df: pd.Dataframe
                Pandas DataFrame object

            feature_name: string
                Specified feature column name.

        Returns:
            Gives back a dictionary object of all possible feature values
            with their associated colors.
        """

        colors = self.__df_features.get_feature_colors(feature_name)
        feature_value_representation = self.__df_features.get_feature_value_representation()

        if colors:
            if isinstance(colors, dict):
                feature_values = df[feature_name].value_counts(
                    sort=False).keys().to_list()
                decoder = self.__df_features.get_label_decoder()

                # Add color feature value for decoders values
                if feature_name in decoder.keys():
                    for cat, val in decoder[feature_name].items():

                        if cat in colors.keys():
                            hex_code = colors[cat]
                            colors[decoder[feature_name][cat]] = hex_code

                        elif val in colors.keys():
                            hex_code = colors[val]
                            colors[cat] = hex_code

                # Add color feature value for different value representation
                if feature_name in feature_value_representation.keys():
                    for val in feature_value_representation[
                        feature_name].keys():
                        if val in colors.keys():
                            hex_code = colors[val]
                            colors[feature_value_representation[
                                feature_name][val]] = hex_code

                i = 0
                for value in feature_values:
                    if value not in colors.keys():
                        colors[value] = \
                        GRAPH_DEFAULTS.DEFINED_LIST_OF_RANDOM_COLORS[i]
                        i += 1

                        if i == len(
                                GRAPH_DEFAULTS.DEFINED_LIST_OF_RANDOM_COLORS):
                            colors = None
        return colors

    def __sort_two_lists(self,
                         sort_values,
                         other_list):
        """

            Sort's two collections by the first collection passed in.

        Args:
            sort_values: collection
                Values to be sorted by.
            other_list: collection
                Values that get sorted based on 'sort_values'.
        Returns:
            Returns back those two lists sorted.
        """
        tmp = list(zip(*sorted(list(zip(other_list, sort_values)),
                               key=lambda x: x[1])))

        return list(tmp[1]), list(tmp[0])


def evaluate_model(self,
                   model_name,
                   model,
                   df,
                   df_features,
                   output_folder=None,
                   le_map=None,
                   show_extra=True,
                   find_nearest_on_cols=False,
                   zscore_low=-2,
                   zscore_high=2):
    """
    model_name:
        The string key to give the dict

    model:
        Cluster model type; it must have '.labels_' as an attribute

    df:
        Dataframe object

    df_features:
        DataFrameTypes object; organizes feature types into groups.

    output_folder:
        Sub directory to put the pngs

    le_map:
        Dict of dataframe cols to LabelEncoders

    show_extra:
        Show extra information from all functions

    find_nearest_on_cols:
            Allows columns to be converted to actual values found within
            the dataset.
            Ex: Can't have .7 in a bool column convert's it to 1.

            False: Just apply to obj columns and bool columns

            True: Apply to all columns

    zscore_low/zscore_high:
        Defines how the threshold to remove data points when profiling the
        cluster.


    The main purpose of 'evaluate_model' is to display/save tables/plots
    accoiated with describing the model's 'findings'.
    """

    df = copy.deepcopy(df)

    # Create folder structure for png outputs
    if not output_folder:
        output_path = str(model).split("(", 1)[0] + "/" + model_name
    else:
        output_path = output_folder + "/" + model_name

    # ---
    # self.__visualize_clusters(model=model,
    #                           output_path=output_path,
    #                           model_name=model_name)

    # ---
    df["Cluster_Name"] = model.labels_
    numerical_features = df_features.numerical_features()
    clustered_dataframes, shrunken_labeled_df = \
        self.__create_cluster_sub_dfs(
            df=df, model=model, numerical_features=numerical_features,
            zscore_low=zscore_low, zscore_high=zscore_high)

    rows_count, cluster_profiles_df = self.__create_cluster_profiles(
        clustered_dataframes=clustered_dataframes,
        shrunken_df=shrunken_labeled_df,
        numerical_features=df_features.numerical_features(),
        le_map=le_map,
        output_path=output_path,
        show=show_extra,
        find_nearest_on_cols=find_nearest_on_cols)

    # Check to see how many data points were removed during the profiling
    # stage
    print("Orginal points in dataframe: ", df.shape[0])
    print("Total points in all modified clusters: ", rows_count)
    print("Shrank by: ", df.shape[0] - rows_count)

    # In case to many data points were removed
    if cluster_profiles_df.shape[0] == 0:
        print(
            "The applied Z-scores caused the cluster profiles "
            "to shrink to far for the model {0}!".format(
                model_name))

    # Display and save dataframe table
    else:
        display(cluster_profiles_df)
        self.__render_mpl_table(cluster_profiles_df, sub_dir=output_path,
                                filename="All_Clusters",
                                header_columns=0, col_width=2.0)

def __create_cluster_profiles(self,
                              clustered_dataframes,
                              shrunken_df,
                              numerical_features,
                              le_map,
                              output_path,
                              find_nearest_on_cols=False,
                              show=True):
    """
        Profile each clustered dataframe based off the given mean.
        Displays extra information in dataframe tables to be understand
        each cluster.

        find_nearest_on_cols:
            Allows columns to be converted to actual values found within
            the dataset.
            Ex: Can't have .7 in a bool column convert's it to 1.

            False: Just apply to obj columns and bool columns

            True: Apply to all columns
    """

    def find_nearest(numbers, target):
        """
            Find the closest fitting number to the target number
        """
        numbers = np.asarray(numbers)
        idx = (np.abs(numbers - target)).argmin()
        return numbers[idx]

    cluster_profiles_df = pd.DataFrame(columns=shrunken_df.columns).drop(
        'Cluster_Name', axis=1)
    rows_count = 0
    for cluster_identfier, cluster_dataframe in \
            clustered_dataframes.items():
        df = pd.DataFrame(columns=cluster_dataframe.columns)
        df = df.append(cluster_dataframe.mean(), ignore_index=True)
        df.index = [cluster_identfier]

        if cluster_dataframe.shape[0] <= 1:
            continue

        # Attempt to convert numbers found within the full set of data
        for col in cluster_dataframe.columns:
            if col not in numerical_features or find_nearest_on_cols:
                df[col] = find_nearest(numbers=shrunken_df[
                    col].value_counts().index.tolist(),
                    target=df[col].values[0])

        # Evaluate cluster dataframe by dataframe
        eval_df = pd.DataFrame(columns=cluster_dataframe.columns)
        eval_df = eval_df.append(
            cluster_dataframe.mean(), ignore_index=True)
        eval_df = eval_df.append(
            cluster_dataframe.min(), ignore_index=True)
        eval_df = eval_df.append(
            cluster_dataframe.median(),
            ignore_index=True)
        eval_df = eval_df.append(
            cluster_dataframe.max(), ignore_index=True)
        eval_df = eval_df.append(
            cluster_dataframe.std(), ignore_index=True)
        eval_df = eval_df.append(
            cluster_dataframe.var(), ignore_index=True)
        eval_df.index = ["Mean", "Min", "Median",
                         "Max", "Standard Deviation", "Variance"]

        if show:
            print("Total found in {0} is {1}".format(
                cluster_identfier, cluster_dataframe.shape[0]))
            self.__render_mpl_table(
                df,
                sub_dir=output_path,
                filename=cluster_identfier +
                "_Means_Rounded_To_Nearest_Real_Numbers",
                header_columns=0,
                col_width=4.0)

            self.__render_mpl_table(
                eval_df,
                sub_dir=output_path,
                filename=cluster_identfier +
                "_Eval_Df",
                header_columns=0,
                col_width=4.0)
            display(df)
            display(eval_df)
            self.__vertical_spacing(7)

        cluster_profiles_df = cluster_profiles_df.append(
            self.__decode_df(df, le_map))

        rows_count += cluster_dataframe.shape[0]

    return rows_count, cluster_profiles_df



from eflow._hidden.parent_objects import DataPipelineSegment

__author__ = "Eric Cacciavillani"
__copyright__ = "Copyright 2019, eFlow"
__credits__ = ["Eric Cacciavillani"]
__license__ = "MIT"
__maintainer__ = "EricCacciavillani"
__email__ = "eric.cacciavillani@gmail.com"

class StringCleaner(DataPipelineSegment):

    def __init__(self,
                 segment_id=None,
                 create_file=True):
        DataPipelineSegment.__init__(self,
                                     object_type=self.__class__.__name__,
                                     segment_id=segment_id,
                                     create_file=create_file)

    def run_widget(self):
        pass

    def filter_to_alphanumeric(self):
        pass

    def filter_to_alphabetical(self):
        pass

    def filter_to_numerical(self):
        pass

    def filter_to_ascii_characters(self):
        pass

    def remove_if_specfic_char_is_found(self):
        pass

    def remove_string(self):
        pass

    def feature_only_should_contain_these_values(self):
        pass



__author__ = "Eric Cacciavillani"
__copyright__ = "Copyright 2019, eFlow"
__credits__ = ["Eric Cacciavillani"]
__license__ = "MIT"
__maintainer__ = "EricCacciavillani"
__email__ = "eric.cacciavillani@gmail.com"

def enum(**enums):
    return type('Enum', (), enums)


SYS_CONSTANTS = enum(PARENT_OUTPUT_FOLDER_NAME="eflow Data")


DEFINED_LIST_OF_RANDOM_COLORS = ["#766604","#61396b","#fe1ce2","#0b03ce","#5b3461",
                                 "#93cff6","#d1ff1c","#3d0126","#1b8535","#241c68",
                                 "#55afc4","#d45261","#0b2547","#20944c","#70a68a",
                                 "#601970","#0e12b5","#8e600d","#af4e47","#114d13",
                                 "#71517b","#5142ce","#52cebd","#a3d833","#1af08f",
                                 "#4ff7e4","#8620f1","#51fae2","#edf3ee","#ae6524"]

GRAPH_DEFAULTS = enum(FIGSIZE=(13, 10),
                      NULL_FIGSIZE=(24, 10),
                      NULL_COLOR="#072F5F",
                      DEFINED_LIST_OF_RANDOM_COLORS=DEFINED_LIST_OF_RANDOM_COLORS)


BOOL_STRINGS = enum(TRUE_STRINGS={"y",
                                  "yes",
                                  "okay",
                                  "t",
                                  "true",
                                  "approve",
                                  "approved",
                                  "accepted",
                                  "alright"},
                    FALSE_STRINGS={"n",
                                   "no",
                                   "f",
                                   "false",
                                   "denined",
                                   "deny",
                                   "refuse",
                                   "abnegate",
                                   "never"})


# FORMATED_STRINGS = enum()
# STRING_CONSTANTS = enum()



from eflow.utils.string_utils import correct_directory_path
from eflow._hidden.custom_exceptions import UnsatisfiedRequirments, SnapshotMismatchError
from eflow.foundation import DataFrameTypes
from eflow.utils.sys_utils import create_dir_structure, \
    dict_to_json_file

import os
import pandas as pd
import math
import copy
import json
import numpy as np

__author__ = "Eric Cacciavillani"
__copyright__ = "Copyright 2019, eFlow"
__credits__ = ["Eric Cacciavillani"]
__license__ = "MIT"
__maintainer__ = "EricCacciavillani"
__email__ = "eric.cacciavillani@gmail.com"


class DataFrameSnapshot:
    """
        Attempts to get a "snapshot" of a dataframe by extracting varying data
        of the pandas dataframe object; then generates a file to later compare
        in a set directory. Helps ensures that a dataset's directory structure
        only ever analyzes/works with that given dataframe.

        Shorthand:
            1: You generate a graph with dataset A and place it in a directory called A's Data.
            2: You generate a graph with dataset B and try to place it in A's Data.

            An error should invoke; thus protecting against creating file's
            that are relating to differing dataframes.

        Note:
            Every function using this object has the ability will have the
            ability to turn on/off this generation and checking.
            'dataframe_snapshot'
    """

    def __init__(self,
                 compare_shape=True,
                 compare_feature_names=True,
                 compare_random_values=True):
        """
        Args:
            compare_shape: bool
                Determines whether or not to create/compare the dataframe's shape for the snapshot

            compare_feature_names: bool
                Determines whether or not to create/compare the dataframe's the feature names for the snapshot


            compare_random_values: bool
                Determines whether or not to create/compare the dataframe's 10 'random'
                values found on each feature. As long as the same dataframe is passed the random values should be the same.
                Note:
                    Will ignore float features because of trailing value problem that all floats have.
        """

        # Copy values
        self.__compare_shape = copy.deepcopy(compare_shape)
        self.__compare_feature_names = copy.deepcopy(compare_feature_names)
        self.__compare_random_values = copy.deepcopy(compare_random_values)

        # Error check; must have at least one boolean
        if not self.__compare_shape and \
                not self.__compare_feature_names and \
                not self.__compare_random_values:
            raise UnsatisfiedRequirments("At least one compare boolean must be "
                                         "set to True for snapshot check to properly work")

    def check_create_snapshot(self,
                              df,
                              df_features,
                              directory_path,
                              sub_dir):
        """

            Compares the passed pandas dataframe object to pre defined json
            file.

        Args:
            df: pd.Dataframe
                Pandas dataframe object.

            directory_path: string
                Output path of the dataset's.

            sub_dir: string
                If set to True than it will visualize the given data.

        Raises:
            Will raise a Mismatch error if the json file didn't match upp with the
            passed dataframe snapshot; causing the program to stop in runtime.
        """
        if not isinstance(df, pd.DataFrame):
            raise TypeError(f"'df' must be a pandas datafram object not a {type(df)}")

        if not isinstance(df_features,DataFrameTypes):
            raise TypeError(f"'df_features' must be a DataFrameTypes object not a {type(df_features)}")


        output_folder_path = create_dir_structure(directory_path,
                                                  sub_dir)

        json_file = output_folder_path + "Dataframe Snapshot.json"

        # Meta Data has already been generated; compare data
        if os.path.isfile(json_file):
            with open(json_file) as file:

                data = json.load(file)
                mismatch_error = None

                # Not using for looping; used for logic breaks
                while True:
                    if self.__compare_shape:
                        list_shape = list(df.shape)
                        if list(data["shape"]) != list_shape:
                            mismatch_error = f'the saved shape {data["shape"]} of the' \
                                             f' dataframe snapshot did not match up with' \
                                             f' the passed dataframe shape {list_shape}.'
                            break

                    # Ensure feature names match up
                    if self.__compare_feature_names:

                        snapshot_features = set(data["feature_names"])
                        passed_features = set(df_features.all_features())

                        feature_difference = snapshot_features.symmetric_difference(
                            passed_features)

                        if feature_difference:
                            mismatch_error = "the following feature name conflicts feature:\n"

                            missing_features = []
                            for feature in feature_difference:
                                if feature in snapshot_features:
                                    missing_features.append(feature)

                            extra_features = []
                            for feature in feature_difference:
                                if feature in passed_features:
                                    extra_features.append(feature)

                            if extra_features:
                                mismatch_error += f"--- Passed dataframe has additional feature(s) than snapshot:\n {extra_features}.\n"

                            if missing_features:
                                mismatch_error += f"--- Passed dataframe is missing the following snapshot feature(s):\n {missing_features}.\n"

                            if extra_features or missing_features:
                                break

                    # Ensure sudo random numbers are chosen again
                    if self.__compare_random_values:
                        compared_data = self.__create_random_values_dict(df,
                                                                         df_features)

                        random_values_matchd_flag = True
                        for k,v in data["random_values"].items():
                            if k in compared_data:
                                if data["random_values"][k] != compared_data[k]:
                                    random_values_matchd_flag = False
                                    break
                        if not random_values_matchd_flag:
                            mismatch_error = f"the 'random' values did not match at feature name '{k}' in the dataframe " \
                                             + "(these 'random' values are based on the shape and name of the column)"
                            break

                    # Break main loop
                    break

                # Error found; raise it
                if mismatch_error is not None:
                    raise SnapshotMismatchError(f"DataFrameSnapshot has raised an error because {mismatch_error}." +
                                                "\nThis error invoked because the directory structure saved a json file "
                                                "containing attributes of the dataframe or a 'snapshot'."
                                                "\nThe given error can be resolved by performing any of the following:"
                                                "\n\t* Pass in the same dataframe as expected."
                                                "\n\t* Disable the snapshot check by changing 'dataframe_snapshot' to False."
                                                "\n\t* Disable save file option by changing the parameter 'save_file' to False."
                                                "\n\t* Or deleting the json object file in the dataset directory under _Extras")

        # JSON file doesn't exist; create file
        else:
            self.__create_dataframe_snapshot_json_file(df,
                                                       output_folder_path)


    def __create_random_values_from_string(self,
                                           feature_name,
                                           hash_type):
        """

            Create a sudo random integer based off the name of the feature and
            a number between 1 and 10.

        Args:
            feature_name: string
                The given feature's string name.

            hash_type: int (1-10)
                Numeric value to determine which.

        Returns:
            Returns back a sudo random value based on arguments.
        """

        # Apply each char to a calculation
        result = 0
        for char_index, char in enumerate(feature_name):
            char = str(char)
            if hash_type == 1:
                result += int(ord(char))
            elif hash_type == 2:
                result += int(ord(char) + 62 * ord(char))
            elif hash_type == 3:
                result += int(ord(char) + 147 * ord(char))
            elif hash_type == 4:
                result += int((ord(char) + 92) * math.pow(ord(char), 3))
            elif hash_type == 5:
                result += int(ord(char) + 49 * math.pow(ord(char), 2))
            elif hash_type == 6:
                result += int((23 + ord(char) + 45) * (3 + ord(char) * 2))
            elif hash_type == 7:
                result += int((ord(char) * 5) + 32 + 8)
            elif hash_type == 8:
                result += int(math.pow(ord(char), 2))
            elif hash_type == 9:
                result += int(ord(char) * 2 + 32 + ord(char) * 2 + 5)
            elif hash_type == 10:
                result += int(ord(char) * 4 * 12 + 76 + math.pow(ord(char), 2))
            else:
                raise ValueError("Hash type must be between 1-10.")

        # Make sure the number is positive (Future proofing if someone adds in
        # something that could cause the result to be negative.)
        result = np.absolute(result)

        return result


    def __create_random_values_dict(self,
                                    df,
                                    df_features):
        """

            Creates a dict of 10 random values for each value based on the
            feature's name and the count of the random values already
            generated.

        Args:
            df: pd.Dataframe
                Pandas dataframe object

            df_features: DataFrameTypes from eflow
                DataFrameTypes object; organizes feature types into groups.

        Returns:
            Returns back a dict of 10 random values based on the arguments.
        """

        feature_values = dict()
        random_indexes = set()
        float_features = set(df_features.float_features())

        # Empty dataframe check
        if not df.shape[0]:
            return feature_values

        for feature in sorted(df_features.all_features()):

            # Ignore if feature is a float
            if feature in float_features:
                continue

            # Random values based on unique indexes for a given feature dict
            feature_values[feature] = dict()
            for hash_type in list(range(1, 11)):
                random_index = self.__create_random_values_from_string(feature,
                                                                       hash_type) % \
                               df.shape[0]
                size_of_indexes = len(random_indexes)

                # Feature needs at least 10 indexes for 10 unique indexes
                if df.shape[0] > 10:
                    # Keep changing the 'random_index' until the value is unique within the set
                    for i in list(range(0, df.shape[0])):

                        random_index += i
                        if random_index >= df.shape[0]:
                            random_index -= df.shape[0]

                        if size_of_indexes != len(random_indexes):
                            break
                else:
                    random_indexes.add(random_index)

                # --------
                feature_values[feature][f'Random Value {hash_type}'] = dict()
                random_value = df[feature].iloc[random_index]

                # Random value is nan
                if not random_value or pd.isnull(random_value):
                    feature_values[feature][
                        f'Random Value {hash_type}'] = "NaN"

                # Convert the value to string no matter what; (Almost all objects have a str representation)
                else:
                    feature_values[feature][
                        f'Random Value {hash_type}'] = str(random_value)

        return feature_values


    def __generate_dataframe_snapshot_dict(self,
                                           df):
        """

            Creates a dict representing a snapshot of the dataframe based on
            the arguments when the object was first inited.

        Args:
            df: pd.Dataframe
                Pandas dataframe object

        Returns:
            Returns back a dict representing the snapshot of the dataframe.

            Note:
                Uses the following private variables to decide whether or not to generate
                that component of the dataframe.
        """
        df_features = DataFrameTypes(df,
                                     ignore_nulls=True)

        meta_dict = dict()
        if self.__compare_shape:
            meta_dict["shape"] = list(df.shape)

        if self.__compare_feature_names:
            meta_dict["feature_names"] = sorted(df_features.all_features())

        if self.__compare_random_values:
            meta_dict["random_values"] = self.__create_random_values_dict(df,
                                                                          df_features)

        return meta_dict


    def __create_dataframe_snapshot_json_file(self,
                                              df,
                                              output_folder_path):
        """

            Creates a json file based on the dataframe's generated snapshot dict.

        Args:
            df: pd.Dataframe
                Pandas Dataframe object

            output_folder_path: string
                Output path the json object will move to.
        """
        output_folder_path = correct_directory_path(output_folder_path)

        meta_dict = self.__generate_dataframe_snapshot_dict(df)

        dict_to_json_file(meta_dict,
                                   output_folder_path,
                                   "Dataframe Snapshot")




import random

__author__ = "Eric Cacciavillani"
__copyright__ = "Copyright 2019, eFlow"
__credits__ = ["Eric Cacciavillani"]
__license__ = "MIT"
__maintainer__ = "EricCacciavillani"
__email__ = "eric.cacciavillani@gmail.com"

def convert_to_filename(filename,
                        remove_file_extension=True):
    """

        Attempts to make the filename string valid.

    Args:
        filename: string
           String of a potential filename.

        remove_file_extension: bool
            Removes everything after the first found value of "." found in the
            string if set to true.

    Returns:
        A string that is valid for saving.
    """
    if remove_file_extension:
        filename = filename.split(".")[0]
    return "".join(x for x in str(
        filename) if (x.isalnum() or x.isascii()) and x != ":")


def correct_directory_path(directory_path):
    """

        Attempts to convert the directory path to a proper one by removing
        any double slashes next to one another.

    Args:
        directory_path:
            String of a potential directory path.

    Returns:
        Returns the fixed path.
    """
    last_char = None
    new_string = ""
    for char in directory_path:
        if last_char and (last_char == "/" and char == "/"):
            pass
        else:
            new_string += char

        last_char = char

    if new_string[-1] != "/":
        new_string += "/"

    return new_string

def create_hex_decimal_string(string_len=10):
    """

        Creates a string of a random Hexadecimal value.

    Args:
        string_len:
            Length of the Hexadecimal string.

    Returns:
        Returns the Hexadecimal string
    """
    return f'%0{string_len}x' % random.randrange(16 ** string_len)


from eflow._hidden.parent_objects import AutoModeler
from eflow.utils.sys_utils import pickle_object_to_file, create_dir_structure, write_object_text_to_file, check_if_directory_exists
from eflow.utils.eflow_utils import move_folder_to_eflow_garbage
from eflow._hidden.custom_exceptions import UnsatisfiedRequirments

# Getting Sklearn Models
from sklearn.decomposition import PCA
from scipy.cluster.hierarchy import linkage, dendrogram,set_link_color_palette
from sklearn.preprocessing import StandardScaler
from sklearn.cluster import AgglomerativeClustering
from sklearn.neighbors import kneighbors_graph

# Getting pyclustering
from pyclustering.cluster.kmeans import kmeans
from pyclustering.cluster.kmedians import kmedians
from pyclustering.cluster.kmedoids import kmedoids
from pyclustering.cluster.ema import ema
from pyclustering.cluster.cure import cure
from pyclustering.cluster.fcm import fcm
from pyclustering.cluster.somsc import somsc
from pyclustering.cluster.center_initializer import kmeans_plusplus_initializer,random_center_initializer

# Visuals libs
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import pylab as pl
import seaborn as sns
from IPython.display import display, HTML

# Misc
from collections import Counter
from scipy.stats import zscore
from kneed import DataGenerator, KneeLocator
import pandas as pd
import six
import random
import numpy as np
import copy
import os
from tqdm import tqdm


__author__ = "Eric Cacciavillani"
__copyright__ = "Copyright 2019, eFlow"
__credits__ = ["Eric Cacciavillani"]
__license__ = "MIT"
__maintainer__ = "EricCacciavillani"
__email__ = "eric.cacciavillani@gmail.com"


class AutoCluster(AutoModeler):
    """
        Analyzes the feature data of a pandas Dataframe object.
    """

    def __init__(self,
                 df,
                 df_features,
                 project_sub_dir="",
                 project_name="Auto Clustering",
                 overwrite_full_path=None,
                 notebook_mode=False,
                 pca_perc=None):
        """
        Args:
            df: pd.Dataframe
                pd.Dataframe

            df_features: Dataframes type holder
                Dataframes type holder

            project_sub_dir: string
                Sub directory to write data.

            project_name: string
                Main project directory

            overwrite_full_path: string
                Overwrite full directory path to a given output folder

            notebook_mode: bool
                Display and show in notebook if set to true.
        """

        AutoModeler.__init__(self,
                             f'{project_sub_dir}/{project_name}',
                             overwrite_full_path)


        if os.path.exists(self.folder_path + "_Extras"):
            move_folder_to_eflow_garbage(self.folder_path+"_Extras",
                                         "Auto Clustering")


        # Define model
        self.__all_cluster_models = dict()

        self.__df_features = copy.deepcopy(df_features)

        self.__notebook_mode = copy.deepcopy(notebook_mode)

        self.__models_suggested_clusters = dict()

        # --- Apply pca ---
        if pca_perc:

            # Create scaler object
            scaler = StandardScaler()
            scaled = scaler.fit_transform(df)

            print("\nInspecting scaled results!")
            self.__inspect_feature_matrix(sub_dir="PCA",
                                          filename="Applied scaler results",
                                          matrix=scaled,
                                          feature_names=df.columns)

            pca, scaled = self.__visualize_pca_variance(scaled)

            # Generate "dummy" feature names
            pca_feature_names = ["PCA_Feature_" +
                                 str(i) for i in range(1,
                                                       len(df.columns) + 1)]

            print("\nInspecting applied scaler and pca results!")
            self.__inspect_feature_matrix(sub_dir="PCA",
                                          filename="Applied scaler and PCA results",
                                          matrix=scaled,
                                          feature_names=pca_feature_names)

            if pca_perc < 1.0:
                # Find cut off point on cumulative sum
                cutoff_index = np.where(
                    pca.explained_variance_ratio_.cumsum() > pca_perc)[0][0]
            else:
                cutoff_index = scaled.shape[1] - 1

            print(
                "After applying pca with a cutoff percentage {0}%"
                " for the cumulative index. Using features 1 to {1}".format(
                    pca_perc, cutoff_index + 1))

            print("Old shape {0}".format(scaled.shape))

            scaled = scaled[:, :cutoff_index + 1]
            pca_feature_names = pca_feature_names[0: cutoff_index + 1]

            print("New shape {0}".format(scaled.shape))

            scaled = scaler.fit_transform(scaled)

            print("\nInspecting data after final scaler applied!")
            self.__inspect_feature_matrix(sub_dir="PCA",
                                          filename="Applied final sclaer to process.",
                                          matrix=scaled,
                                          feature_names=pca_feature_names)

            self.__scaled = scaled

        # Assumed PCA has already been applied; pass as matrix
        else:
            self.__scaled = df.values


    # --- Getters/Setters
    def get_scaled_data(self):
        """

            Gets the stored data

        Returns:
            Returns the stored data
        """

        return copy.deepcopy(self.__scaled)

    def get_all_cluster_models(self):
        """

            Gets the model names and model instances in dictionary form.

        Return:
            Returns the model name to model instance dict
        """
        return copy.deepcopy(self.__all_cluster_models)

    def delete_stored_data(self):
        """

            Removes the matrix data in order to save RAM when running
            analysis on the actual data.
        """
        del self.__scaled
        self.__scaled = None

    def visualize_hierarchical_clustering(self,
                                          linkage_methods=None,
                                          display_print=True,
                                          display_visuals=True):
        """

            Displays hierarchical cluster graphs with provided methods.

        Args:
            linkage_methods:
                All methods applied to the linkage

            display_print:
                Display print outputs

            display_visuals:
                Display plot data
        """

        best_clusters = []

        # ---
        if not linkage_methods:
            linkage_methods = ["complete",
                               "single",
                               "weighted",
                               "ward",
                               "average",
                               "centroid",
                               "median"]

        # Apply methods to each dendrogram
        for method in linkage_methods:

            if display_print:
                print(f"Creating graphic for Hierarchical Clustering Method: {method}...")

            # Create mergings
            mergings = linkage(self.__scaled, method=method)

             # {"Yellow":"#d3d255",
             # "Magenta":"#c82bc9",
             # "Black":"#030303",
             # "Red":"#ff403e",
             # "Green":"#3f9f3f",
             # "Cyan":"#0ec1c2",
             # "Brown": "#775549",
             # "Silver": "#C0C0C0",
             # "Blue":"#24326f",
             # "Orange":"#cc7722",
             # "Mauve":"#9c7c8c"}

            # Set plot
            plt.figure(figsize=(12, 7))
            set_link_color_palette(None)

            # Plot the dendrogram, using varieties as labels
            color_list = dendrogram(mergings,
                                    labels=list(range(0, len(self.__scaled,))),
                                    leaf_rotation=90,
                                    leaf_font_size=3)["color_list"]

            method = method.capitalize()
            plt.title(f"Hierarchical Clustering Method : \'{method}\'")

            # -----
            self.save_plot("Hierarchical Clustering",
                           f"Hierarchical Clustering Method {method} without legend")

            del mergings


            # Create proper cluster names based on the color names
            color_cluster_count = dict()
            last_color = None
            known_colors = set()
            color_cluster_order = list()
            seq_len = 0
            i = 0

            # -----
            for color in copy.deepcopy(color_list):

                # Proper color name
                color = self.__get_color_name(color)

                # Name for old cluster color sequence found
                if color in known_colors:
                    color_list[i] = f"{color} cluster {color_cluster_count[color]}"

                # Name for new cluster color sequence found
                else:
                    color_list[i] = f"{color} cluster 0"

                # Track unique cluster color order
                if color_list[i] not in color_cluster_order:
                    color_cluster_order.append(color_list[i])

                if last_color:
                    # Sequence of color has yet to be broken
                    if last_color == color:

                        # Only really need to check if the sequence has a length of 1
                        if seq_len <= 2:
                            seq_len += 1

                    # Sequence break
                    else:

                        # Valid cluster found
                        if seq_len > 1:

                            # Track all known color names
                            if last_color not in known_colors:
                                known_colors.add(last_color)

                            # If color is repeated; then make a new cluster count name
                            if last_color not in color_cluster_count.keys():
                                color_cluster_count[last_color] = 1
                            else:
                                color_cluster_count[last_color] += 1

                        # Invalid color cluster found; apply needed changes
                        else:
                            color_list.pop(i-1)
                            i -= 1

                        seq_len = 0

                last_color = color
                i += 1

            # Create legend for each cluster name and the amount of per sample
            counter_object = Counter(color_list)
            cluster_color_count = dict()
            handles = []

            # Make sure the legend is in the same order the cluster appear in the dendrogram
            for color_cluster_name in color_cluster_order:
                if color_cluster_name in counter_object.keys():
                    cluster_color_count[color_cluster_name] = counter_object[color_cluster_name]
                    try:
                        handles.append(mpatches.Patch(
                            color=color_cluster_name.split(" cluster ")[0],
                            label=color_cluster_name + f": {counter_object[color_cluster_name]} samples"))

                    except:
                        handles.append(mpatches.Patch(
                            color="black",
                            label=color_cluster_name + f": {counter_object[color_cluster_name]} samples"))

            # Plot the legend and save the plot
            plt.legend(handles=handles,
                       loc='upper right',
                       bbox_to_anchor=(1.32, 1.01),
                       title=f"Clusters ({len(handles)})")

            best_clusters.append(len(handles))
            if display_visuals and self.__notebook_mode:
                plt.show()

            self.save_plot("Hierarchical Clustering",
                           f"Hierarchical Clustering Method {method} with legend")

            plt.close('all')

        # Save results into _Extras folder
        best_clusters.sort()
        self.__models_suggested_clusters["Hierarchical Clustering"] = best_clusters
        self.__save_update_best_model_clusters()
        return best_clusters


    # def __visualize_clusters(self, model, output_path, model_name=""):
    #     """
    #         Creates visualization of clustering model on given data.
    #     """
    #     markers = ["+", "*", "X", "o", "v", "P", "H", "4", "p", "D", "s",
    #                "1", "x", "d", "_"]
    #     colors = ['b', 'g', 'r', 'c', 'm', 'y',
    #               '#007BA7', '#ff69b4', '#CD5C5C', '#7eab19', '#1a4572',
    #               '#2F4F4F', '#4B0082', '#d11141', '#5b2504']
    #
    #     # Display ranking on color based on amount data points per cluster
    #     unique, counts = np.unique(model.labels_, return_counts=True)
    #     cluster_names = ["Cluster:" + str(cluster_label)
    #                      for cluster_label in unique]
    #     self.__display_rank_graph(feature_names=cluster_names,
    #                               metric=counts,
    #                               title=model_name,
    #                               output_path=output_path,
    #                               model_name=model_name,
    #                               y_title="Clusters",
    #                               x_title="Found per cluster")
    #     pl.figure(figsize=(8, 7))
    #
    #     # Display clustered graph
    #     cluster_array = list(range(0, len(cluster_names)))
    #     scaled_cluster_label = np.hstack(
    #         (self.__scaled, np.reshape(
    #             model.labels_.astype(int), (self.__scaled.shape[0], 1))))
    #     for i in range(0, scaled_cluster_label.shape[0]):
    #         cluster_label = int(scaled_cluster_label[i][-1])
    #         cluster_array[cluster_label] = pl.scatter(
    #             scaled_cluster_label[i, 0], scaled_cluster_label[i, 1],
    #             c=colors[cluster_label], marker=str(markers[cluster_label]))
    #
    #     pl.legend(cluster_array, cluster_names)
    #     pl.title(model_name + ' visualized with data', fontsize=15)
    #     self.__image_processing_utils(output_path,
    #                           model_name + "_Visualized_Cluster")
    #     plt.show()
    #     plt.close()
    #     pl.close()
    def create_elbow_models(self,
                            model_names=["K-Means",
                                         "K-Medians",
                                         "K-Medoids",
                                         "Somsc",
                                         "Cure",
                                         "Fuzzy C-means"],
                            repeat_operation=3,
                            max_k_value=15,
                            display_visuals=True):

        model_names = set(model_names)

        names_model_dict = {"K-Means":kmeans,
                            "K-Medians":kmedians,
                            "K-Medoids":kmedoids,
                            "Somsc":somsc,
                            "Cure":cure,
                            "Fuzzy C-means": fcm}

        # Iterate through passed model names
        for name in model_names:

            if name in names_model_dict.keys():

                # Only requires 1 elbow sequence
                if name == "Somsc" or name == "Cure":
                    best_clusters = self.__create_elbow_seq(name,
                                                            names_model_dict[name],
                                                            repeat_operation=1,
                                                            max_k_value=max_k_value,
                                                            display_visuals=display_visuals)
                else:
                    best_clusters = self.__create_elbow_seq(name,
                                                            names_model_dict[name],
                                                            repeat_operation=repeat_operation,
                                                            max_k_value=max_k_value,
                                                            display_visuals=display_visuals)

                # Save cluster results in
                best_clusters.sort()
                self.__models_suggested_clusters[name] = best_clusters
                self.__save_update_best_model_clusters()
            else:
                raise UnsatisfiedRequirments(f"Unknown model name passed: \"{name}\"")

        return best_clusters

    def create_agglomerative_models(self,
                                    n_cluster_list,
                                    linkage_methods=None):
        """
            Create multiple agglomerative models based on a list of
            'n_clusters' values and defined linkage methods.
        """

        if isinstance(n_cluster_list, int):
            n_cluster_list = [n_cluster_list]

        if not linkage_methods:
            linkage_methods = ["ward", "complete", "average", "single"]

        knn_graph = kneighbors_graph(
            self.__scaled, len(
                self.__scaled) - 1, include_self=False)

        for n_clusters in n_cluster_list:
            for connectivity in (None, knn_graph):

                for _, linkage in enumerate(linkage_methods):
                    model = AgglomerativeClustering(linkage=linkage,
                                                    connectivity=connectivity,
                                                    n_clusters=n_clusters)
                    model.fit(self.__scaled)
                    self.__all_cluster_models[
                        "AgglomerativeClustering_{0}_"
                        "cluster{1}_Connectivity{2}".format(
                            linkage,
                            n_clusters, connectivity is not None)] = model

                    print(
                        "Successfully generate Agglomerative model with "
                        "linkage {0} on n_clusters={1}".format(
                            linkage, n_clusters))

        # self.__models_suggested_clusters["Agglomerative models"] =

    def evaluate_all_models(self,
                            df,
                            df_features,
                            le_map=None,
                            show_extra=True,
                            find_nearest_on_cols=False,
                            zscore_low=-2,
                            zscore_high=2):
        """
            Loop through all models and evaluate the given model with
            'evaluate_model'. Read 'evaluate_model' to learn more.
        """

        for model_name, model in self.__all_cluster_models.items():
            self.evaluate_model(model_name=model_name,
                                model=model,
                                df=df,
                                df_features=df_features,
                                le_map=le_map,
                                find_nearest_on_cols=find_nearest_on_cols,
                                show_extra=show_extra,
                                zscore_low=zscore_low,
                                zscore_high=zscore_high)

            self.__vertical_spacing(5)
            print("----" * 20)

    def evaluate_model(self,
                       model_name,
                       model,
                       df,
                       df_features,
                       output_folder=None,
                       le_map=None,
                       show_extra=True,
                       find_nearest_on_cols=False,
                       zscore_low=-2,
                       zscore_high=2):
        """
        model_name:
            The string key to give the dict

        model:
            Cluster model type; it must have '.labels_' as an attribute

        df:
            Dataframe object

        df_features:
            DataFrameTypes object; organizes feature types into groups.

        output_folder:
            Sub directory to put the pngs

        le_map:
            Dict of dataframe cols to LabelEncoders

        show_extra:
            Show extra information from all functions

        find_nearest_on_cols:
                Allows columns to be converted to actual values found within
                the dataset.
                Ex: Can't have .7 in a bool column convert's it to 1.

                False: Just apply to obj columns and bool columns

                True: Apply to all columns

        zscore_low/zscore_high:
            Defines how the threshold to remove data points when profiling the
            cluster.


        The main purpose of 'evaluate_model' is to display/save tables/plots
        accoiated with describing the model's 'findings'.
        """

        df = copy.deepcopy(df)

        # Create folder structure for png outputs
        if not output_folder:
            output_path = str(model).split("(", 1)[0] + "/" + model_name
        else:
            output_path = output_folder + "/" + model_name

        # ---
        # self.__visualize_clusters(model=model,
        #                           output_path=output_path,
        #                           model_name=model_name)

        # ---
        df["Cluster_Name"] = model.labels_
        numerical_features = df_features.numerical_features()
        clustered_dataframes, shrunken_labeled_df = \
            self.__create_cluster_sub_dfs(
                df=df, model=model, numerical_features=numerical_features,
                zscore_low=zscore_low, zscore_high=zscore_high)

        rows_count, cluster_profiles_df = self.__create_cluster_profiles(
            clustered_dataframes=clustered_dataframes,
            shrunken_df=shrunken_labeled_df,
            numerical_features=df_features.numerical_features(),
            le_map=le_map,
            output_path=output_path,
            show=show_extra,
            find_nearest_on_cols=find_nearest_on_cols)

        # Check to see how many data points were removed during the profiling
        # stage
        print("Orginal points in dataframe: ", df.shape[0])
        print("Total points in all modified clusters: ", rows_count)
        print("Shrank by: ", df.shape[0] - rows_count)

        # In case to many data points were removed
        if cluster_profiles_df.shape[0] == 0:
            print(
                "The applied Z-scores caused the cluster profiles "
                "to shrink to far for the model {0}!".format(
                    model_name))

        # Display and save dataframe table
        else:
            display(cluster_profiles_df)
            self.__render_mpl_table(cluster_profiles_df, sub_dir=output_path,
                                    filename="All_Clusters",
                                    header_columns=0, col_width=2.0)

    def __create_cluster_profiles(self,
                                  clustered_dataframes,
                                  shrunken_df,
                                  numerical_features,
                                  le_map,
                                  output_path,
                                  find_nearest_on_cols=False,
                                  show=True):
        """
            Profile each clustered dataframe based off the given mean.
            Displays extra information in dataframe tables to be understand
            each cluster.

            find_nearest_on_cols:
                Allows columns to be converted to actual values found within
                the dataset.
                Ex: Can't have .7 in a bool column convert's it to 1.

                False: Just apply to obj columns and bool columns

                True: Apply to all columns
        """

        def find_nearest(numbers, target):
            """
                Find the closest fitting number to the target number
            """
            numbers = np.asarray(numbers)
            idx = (np.abs(numbers - target)).argmin()
            return numbers[idx]

        cluster_profiles_df = pd.DataFrame(columns=shrunken_df.columns).drop(
            'Cluster_Name', axis=1)
        rows_count = 0
        for cluster_identfier, cluster_dataframe in \
                clustered_dataframes.items():
            df = pd.DataFrame(columns=cluster_dataframe.columns)
            df = df.append(cluster_dataframe.mean(), ignore_index=True)
            df.index = [cluster_identfier]

            if cluster_dataframe.shape[0] <= 1:
                continue

            # Attempt to convert numbers found within the full set of data
            for col in cluster_dataframe.columns:
                if col not in numerical_features or find_nearest_on_cols:
                    df[col] = find_nearest(numbers=shrunken_df[
                        col].value_counts().index.tolist(),
                        target=df[col].values[0])

            # Evaluate cluster dataframe by dataframe
            eval_df = pd.DataFrame(columns=cluster_dataframe.columns)
            eval_df = eval_df.append(
                cluster_dataframe.mean(), ignore_index=True)
            eval_df = eval_df.append(
                cluster_dataframe.min(), ignore_index=True)
            eval_df = eval_df.append(
                cluster_dataframe.median(),
                ignore_index=True)
            eval_df = eval_df.append(
                cluster_dataframe.max(), ignore_index=True)
            eval_df = eval_df.append(
                cluster_dataframe.std(), ignore_index=True)
            eval_df = eval_df.append(
                cluster_dataframe.var(), ignore_index=True)
            eval_df.index = ["Mean", "Min", "Median",
                             "Max", "Standard Deviation", "Variance"]

            if show:
                print("Total found in {0} is {1}".format(
                    cluster_identfier, cluster_dataframe.shape[0]))
                self.__render_mpl_table(
                    df,
                    sub_dir=output_path,
                    filename=cluster_identfier +
                    "_Means_Rounded_To_Nearest_Real_Numbers",
                    header_columns=0,
                    col_width=4.0)

                self.__render_mpl_table(
                    eval_df,
                    sub_dir=output_path,
                    filename=cluster_identfier +
                    "_Eval_Df",
                    header_columns=0,
                    col_width=4.0)
                display(df)
                display(eval_df)
                self.__vertical_spacing(7)

            cluster_profiles_df = cluster_profiles_df.append(
                self.__decode_df(df, le_map))

            rows_count += cluster_dataframe.shape[0]

        return rows_count, cluster_profiles_df

    def __create_cluster_sub_dfs(self,
                                 df,
                                 model,
                                 numerical_features,
                                 zscore_low=-2,
                                 zscore_high=2):
        """
            Shrinks the clustered dataframe by rows based on outliers
            found within each cluster.

            Returns back a dict of dataframes with specficed clustered values
            alongside a full dataframe comprised of those clustered dataframes.
        """
        # Dataframe to analyze model 'better' choices
        shrunken_full_df = df.drop('Cluster_Name', axis=1).drop(df.index)

        # Store each sub-dataframe based on cluster label
        clustered_dataframes = dict()

        for cluster_label in set(model.labels_):
            cluster_df = df[df["Cluster_Name"] == cluster_label]
            # Ignore cluster with only one patient
            if len(cluster_df) <= 1:
                continue
            # ---
            zscore_cluster_df = cluster_df.drop(
                'Cluster_Name', axis=1).apply(zscore)

            # Check if cluster is only comprised of one data point
            if cluster_df.shape[0] > 1:

                # Iterate through all numerical features
                for numerical_feature in numerical_features:

                    nan_check = zscore_cluster_df[
                        numerical_feature].isnull().values.any()
                    # Check for nans
                    if not nan_check:
                        zscore_cluster_df = zscore_cluster_df[
                            zscore_cluster_df[numerical_feature] >= zscore_low]
                        zscore_cluster_df = zscore_cluster_df[
                            zscore_cluster_df[numerical_feature] <= zscore_high]

            # Dummy list of -1s alloc at given pos of 'zscore_cluster_df'
            # indexes
            reshaped_index = [-1] * len(df.index.values)

            for given_index in list(zscore_cluster_df.index.values):
                reshaped_index[given_index] = given_index

            # Pass back all vectors that passed the zscore test
            bool_array = pd.Series(reshaped_index).astype(int) == pd.Series(
                list(df.index.values)).astype(int)

            temp_cluster_df = df[bool_array].reset_index(drop=True)

            # Store in proper collection objs
            shrunken_full_df = shrunken_full_df.append(temp_cluster_df)

            clustered_dataframes[
                "Cluster:" + str(cluster_label)] = temp_cluster_df.drop(
                'Cluster_Name', axis=1)

        return clustered_dataframes, shrunken_full_df

    def __inspect_feature_matrix(self,
                                 sub_dir,
                                 filename,
                                 matrix,
                                 feature_names):
        """
            Creates a dataframe to quickly analyze a matrix
        """
        mean_matrix = np.mean(matrix, axis=0)
        std_matrix = np.std(matrix, axis=0)
        data_dict = dict()
        for index, feature_name in enumerate(feature_names):
            data_dict[feature_name] = [mean_matrix[index],
                                       std_matrix[index]]

        tmp_df = pd.DataFrame.from_dict(data_dict,
                                        orient='index',
                                        columns=['Mean', 'Standard Dev'])

        if self.__notebook_mode:
            display(tmp_df)
        else:
            print(tmp_df)

        self.save_table_as_plot(tmp_df,
                                sub_dir=sub_dir,
                                filename=filename,
                                show_index=True,
                                format_float_pos=5)

    # Not created by me!
    # Created by my teacher: Narine Hall
    def __visualize_pca_variance(self, data):
        """
            Visualize PCA matrix feature importance
        """

        # Check for pca variance
        pca = PCA()
        data = pca.fit_transform(data)

        # ----
        plt.bar(range(1, len(pca.explained_variance_ratio_) + 1),
                pca.explained_variance_ratio_)
        plt.xticks()
        plt.ylabel('Variance ratio')
        plt.xlabel('PCA feature')
        plt.tight_layout()

        self.save_plot("PCA",
                       "PCA Feature Variance Ratio")

        if self.__notebook_mode:
            plt.show()
            plt.close("all")

        # ----
        plt.bar(range(1, len(pca.explained_variance_ratio_) + 1),
                pca.explained_variance_ratio_.cumsum())
        plt.xticks()
        plt.ylabel('Cumulative sum of variances')
        plt.xlabel('PCA feature')
        plt.tight_layout()
        self.save_plot("PCA",
                       "PCA Cumulative Sum of Variances")

        if self.__notebook_mode:
            plt.show()
            plt.close("all")

        return pca, data

    def __display_rank_graph(self,
                             feature_names,
                             metric,
                             output_path,
                             model_name,
                             title="",
                             y_title="",
                             x_title="", ):
        """
            Darker colors have higher rankings (values)
        """
        plt.figure(figsize=(7, 7))

        # Init color ranking fo plot
        # Ref: http://tinyurl.com/ydgjtmty
        pal = sns.color_palette("GnBu_d", len(metric))
        rank = np.array(metric).argsort().argsort()
        ax = sns.barplot(y=feature_names, x=metric,
                         palette=np.array(pal[::-1])[rank])
        plt.xticks(rotation=0, fontsize=15)
        plt.yticks(fontsize=15)
        plt.xlabel(x_title, fontsize=20, labelpad=20)
        plt.ylabel(y_title, fontsize=20, labelpad=20)
        plt.title(title, fontsize=15)
        plt.show()

    def __get_color_name(self,
                         color):

        if color == "b":
            return "Blue"

        elif color == "g":
            return "Green"

        elif color == "r":
            return "Red"

        elif color == "c":
            return "Cyan"

        elif color == "m":
            return "Magenta"

        elif color == "y":
            return "Yellow"

        elif color == "k":
            return "Black"

        elif color == "w":
            return "White"

        else:
            return color

    def __nearest(self, clusters, x):
        return np.argmin([self.__distance(x, c) for c in clusters])

    def __distance(self, a, b):
        return np.sqrt(((a - b) ** 2).sum())

    def __get_unique_random_indexes(self,
                                    k_val):
        random_indexes = set()
        while len(random_indexes) != k_val:

            index = random.randint(0, len(self.__scaled) - 1)

            if index not in random_indexes:
                random_indexes.add(index)

        return random_indexes


    def __create_elbow_seq(self,
                           model_name,
                           model_instance,
                           repeat_operation,
                           max_k_value,
                           display_visuals):
        """
            Generate models based on the found 'elbow' of the interia values.
        """

        max_k_value += 1

        k_models = []
        inertias = []

        for elbow_seq_count in range(0,repeat_operation):
            tmp_inertias = []
            tmp_k_models = []

            if display_visuals:
                pbar = tqdm(range(1,max_k_value), desc=f"{model_name} Elbow Seq Count {elbow_seq_count + 1}")
            else:
                pbar = range(1,max_k_value)

            for k_val in pbar:

                if display_visuals:
                    pbar.set_postfix(model_count=k_val, refresh=True)


                model = self.__create_pyclustering_model(model_name=model_name,
                                                         model_instance=model_instance,
                                                         k_val=k_val)

                # Run cluster analysis and obtain results.
                model.process()
                final_centers = np.array(self.__get_centers(model))
                labels = [self.__nearest(final_centers, x) for x in self.__scaled]

                inertia = sum(((final_centers[l] - x) ** 2).sum()
                              for x, l in zip(self.__scaled, labels))

                # Append the inertia to the list of inertias
                tmp_inertias.append(inertia)
                tmp_k_models.append(model)

            k_models.append(tmp_k_models)
            inertias.append(tmp_inertias)

        return self.__find_best_elbow_models(model_name,
                                             k_models,
                                             inertias,
                                             display_visuals)

    def __get_centers(self,
                      model):
        try:
            return model.get_centers()
        except AttributeError:

            center_points = []

            for cluster_indexes in model.get_clusters():
                all_dps = np.matrix([self.__scaled[i] for i in cluster_indexes])
                center_dp = all_dps.mean(0)

                # Grave Yard code: Use existing point rather than generating abstract average data point
                # np.absolute(all_dps - center_dp).sum(1).argmin()

                center_points.append(np.array(center_dp.tolist()[0]))

            return center_points


    def __create_pyclustering_model(self,
                                    model_name,
                                    model_instance,
                                    k_val):
        if model_name == "K-Medoids":
            model = model_instance(self.__scaled,
                                       [i for i in
                                        self.__get_unique_random_indexes(
                                            k_val)])
        elif model_name == "Somsc" or model_name == "Cure":
            model = model_instance(self.__scaled,
                               k_val)

        elif model_name == "K-Means" or model_name == "Fuzzy C-means":
            initial_centers = kmeans_plusplus_initializer(self.__scaled, k_val).initialize()
            model = model_instance(self.__scaled, initial_centers)

        else:
            # Create instance of K-Means algorithm with prepared centers.
            initial_centers = random_center_initializer(self.__scaled,
                                                        k_val).initialize()
            model = model_instance(self.__scaled, initial_centers)

        return model




    def __find_best_elbow_models(self,
                                 model_name,
                                 k_models,
                                 inertias,
                                 display_visuals=True):

        ks = range(1, len(inertias[0]) + 1)

        plt.figure(figsize=(13, 6))
        plt.title(f"All possible {model_name} Elbow's", fontsize=15)
        plt.xlabel('Number of clusters, k')
        plt.ylabel('Inertia')
        plt.xticks(ks)

        elbow_inertias_matrix = None
        inertias_matrix = None
        elbow_models = []
        elbow_sections = []
        center_elbow_count = dict()
        proximity_elbow_count = dict()

        # Plot ks vs inertias
        for i in range(0,len(inertias)):

            elbow_cluster = KneeLocator(ks,
                                        inertias[i],
                                        curve='convex',
                                        direction='decreasing').knee

            if elbow_cluster == 1 or not elbow_cluster:
                print("Elbow was either one or None for the elbow seq.")
                continue

            plt.plot(ks,
                     inertias[i],
                     '-o',
                     color='#367588',
                     alpha=0.5)

            if str(elbow_cluster) not in center_elbow_count.keys():
                center_elbow_count[str(elbow_cluster)] = 1
            else:
                center_elbow_count[str(elbow_cluster)] += 1

            for k_val in [elbow_cluster - 1, elbow_cluster, elbow_cluster + 1]:
                elbow_sections.append([ks[k_val - 1],inertias[i][k_val - 1]])

                if str(k_val) not in proximity_elbow_count.keys():
                    proximity_elbow_count[str(k_val)] = 1
                else:
                    proximity_elbow_count[str(k_val)] += 1


            if isinstance(elbow_inertias_matrix, type(None)):
                inertias_matrix = np.matrix(inertias[i])
                elbow_inertias_matrix = np.matrix(inertias[i][elbow_cluster - 2:elbow_cluster + 1])

            else:
                inertias_matrix = np.vstack([inertias_matrix, inertias[i]])

                elbow_inertias_matrix = np.vstack(
                    [elbow_inertias_matrix, inertias[i][elbow_cluster - 2:elbow_cluster + 1]])

            elbow_models.append(k_models[i][elbow_cluster - 2:elbow_cluster + 1])

        for elbow in elbow_sections:
            k_val = elbow[0]
            intertia = elbow[1]
            plt.plot(k_val,
                     intertia,
                     'r*',)

        del inertias
        del k_models
        del elbow_cluster

        self.save_plot(f"Models/{model_name}",f"All possible {model_name} Elbow's",)

        if display_visuals and self.__notebook_mode:
            plt.show()
        plt.close("all")

        center_elbow_count = pd.DataFrame({"Main Knees": list(center_elbow_count.keys()),
                                           "Counts": list(center_elbow_count.values())})
        center_elbow_count.sort_values(by=['Counts'],
                                       ascending=False,
                                       inplace=True)

        self.save_table_as_plot(
            center_elbow_count,
            sub_dir=f"Models/{model_name}",
            filename="Center Elbow Count")

        proximity_elbow_count = pd.DataFrame({"Proximity Knees": list(proximity_elbow_count.keys()),
                                              "Counts": list(proximity_elbow_count.values())})
        proximity_elbow_count.sort_values(by=['Counts'],
                                          ascending=False,
                                          inplace=True)

        self.save_table_as_plot(
            proximity_elbow_count,
            sub_dir=f"Models/{model_name}",
            filename="Proximity Elbow Count")

        plt.figure(figsize=(13, 6))
        plt.title(f"Best of all {model_name} Elbows", fontsize=15)
        plt.xlabel('Number of clusters, k')
        plt.ylabel('Inertia')
        plt.xticks(ks)

        average_elbow_inertias = elbow_inertias_matrix.mean(0)

        knee_vote = []
        for vector in elbow_inertias_matrix:
            knee_vote.append(
                np.absolute(vector - average_elbow_inertias).sum())

        best_elbow_index = np.array(knee_vote).argmin()

        plt.plot(ks,
                 inertias_matrix[best_elbow_index].tolist()[0],
                 '-o',
                 color='#367588')

        best_clusters = []
        for model in elbow_models[best_elbow_index]:
            k_val = len(model.get_clusters())

            self.__all_cluster_models[f"{model_name}_Cluster_" + str(k_val)] = model

            create_dir_structure(self.folder_path,
                                 f"Models/{model_name}/Clusters={k_val}")

            try:
                pickle_object_to_file(model,
                                      self.folder_path + f"Models/{model_name}/Clusters={k_val}",
                                      f"{model_name}_Cluster_" + str(k_val))
            except:
                print(f"Something went wrong when trying to save the model: {model_name}")
            plt.plot(ks[k_val - 1],
                     inertias_matrix[best_elbow_index].tolist()[0][k_val - 1],
                     'r*')
            best_clusters.append(k_val)

        self.save_plot(f"Models/{model_name}",
                       f"Best of all {model_name} Elbows")

        if display_visuals and self.__notebook_mode:
            plt.show()
        plt.close("all")

        best_clusters.sort()

        if display_visuals and self.__notebook_mode:
            display(proximity_elbow_count)
            display(center_elbow_count)

        return best_clusters

    def __save_update_best_model_clusters(self):

        create_dir_structure(self.folder_path,
                             "_Extras")

        pickle_object_to_file(self.__models_suggested_clusters,
                              self.folder_path + "_Extras",
                              "All suggested clusters")
        write_object_text_to_file(self.__models_suggested_clusters,
                                  self.folder_path + "_Extras",
                                  "All suggested clusters")

        all_clusters = []
        for model_name, best_clusters in self.__models_suggested_clusters.items():
            write_object_text_to_file(best_clusters,
                                      self.folder_path + "_Extras",
                                      f"{model_name} suggested clusters")

            all_clusters += best_clusters


        write_object_text_to_file(round(sum(all_clusters) / len(all_clusters)),
                                  self.folder_path + "_Extras",
                                  "Average of suggested clusters")





